# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import sys

from .console import log
from . import util

# TODO: Some verification of the config values


class Config(object):
    """
    Manages the configuration for a benchmark project.
    """
    api_version = 1

    def __init__(self):
        self.project = "project"
        self.project_url = "#"
        self.repo = None
        self.repo_subdir = ""
        self.branches = [None]
        self.pythons = ["{0[0]}.{0[1]}".format(sys.version_info)]
        self.matrix = {}
        self.exclude = []
        self.include = []
        self.env_dir = "env"
        self.benchmark_dir = "benchmarks"
        self.results_dir = "results"
        self.html_dir = "html"
        self.show_commit_url = "#"
        self.hash_length = 8
        self.environment_type = None
        self.install_timeout = 600
        self.dvcs = None
        self.regressions_first_commits = {}
        self.regressions_thresholds = {}
        self.plugins = []
        self.conda_channels = []
        self.conda_environment_file = None
        self.build_command = None
        self.install_command = None
        self.uninstall_command = None

    @classmethod
    def load(cls, path=None):
        """
        Load a configuration from a file.  If no file is provided,
        defaults to `asv.conf.json`.
        """
        if not path:
            path = "asv.conf.json"

        if not os.path.isfile(path):
            raise util.UserError("Config file {0} not found.".format(path))

        d = util.load_json(path, cls.api_version, js_comments=True)
        try:
            return cls.from_json(d)
        except ValueError:
            raise util.UserError(
                "No repo specified in {0} config file.".format(path))

    @classmethod
    def from_json(cls, d):
        if 'wheel_cache_size' in d:
            log.warning("`wheel_cache_size` has been renamed to `build_cache_size`. Update your `asv.conf.json` accordingly.")
            d.setdefault('build_cache_size', d['wheel_cache_size'])

        conf = cls()
        conf.__dict__.update(d)

        if not getattr(conf, "repo", None):
            raise util.UserError(
                "No repo specified in config file.")

        if not getattr(conf, "branches", [None]):
            # If 'branches' attribute is present, at least some must
            # be listed.
            raise util.UserError(
                "No branches specified in config file.")

        return conf



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, unicode_literals, print_function

from ..console import log
from ..publishing import OutputPublisher


class SummaryGrid(OutputPublisher):
    name = ""
    button_label = "Grid view"
    description = "Display as a agrid"
    order = 0

    @classmethod
    def publish(cls, conf, repo, benchmarks, graphs, revisions):
        # Generate and save summary graphs
        summaries = graphs.get_summary_graphs(dots=log.dot)
        for graph in summaries:
            graph.save(conf.html_dir)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)


from .. import profiling
from .. import util


class KCachegrindGui(profiling.ProfilerGui):
    name = 'kcachegrind'
    description = "kcachegrind through pyprof2calltree"

    @classmethod
    def is_available(cls):
        return (
            util.has_command("kcachegrind") and
            util.has_command("pyprof2calltree"))

    @classmethod
    def open_profiler_gui(cls, profiler_file):
        command = util.which("pyprof2calltree")

        return util.check_call(
            [command, '-i', profiler_file, '-k'], timeout=None)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, unicode_literals, print_function

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)


from .. import profiling
from .. import util


class SnakevizGui(profiling.ProfilerGui):
    name = 'snakeviz'
    description = "snakeviz http://jiffyclub.github.io/snakeviz/"

    @classmethod
    def is_available(cls):
        return util.has_command('snakeviz')

    @classmethod
    def open_profiler_gui(cls, profiler_file):
        command = util.which('snakeviz')

        return util.check_call(
            [command, profiler_file],
            valid_return_codes=(0, -15), timeout=None)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import io
import json
import os
import sys
import re
import time
import tempfile
import itertools
import datetime
import pstats
import socket
import struct
import threading
import traceback

import six

from .console import log
from .results import Results, format_benchmark_result
from . import statistics
from . import util


WIN = (os.name == "nt")


# Can't use benchmark.__file__, because that points to the compiled
# file, so it can't be run by another version of Python.
BENCHMARK_RUN_SCRIPT = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "benchmark.py")


JSON_ERROR_RETCODE = -257


BenchmarkResult = util.namedtuple_with_doc(
    'BenchmarkResult',
    ['result', 'samples', 'number', 'errcode', 'stderr', 'profile'],
    """
    Postprocessed benchmark result

    Attributes
    ----------
    result : list of object
        List of numeric values of the benchmarks (one for each parameter
        combination).
        Values are `None` if benchmark failed or NaN if it was skipped.
    samples : list of {list, None}
        List of lists of sampled raw data points (or Nones if
        no sampling done).
    number : list of {dict, None}
        List of actual repeat counts for each sample (or Nones if
        no sampling done).
    errcode : int
        Process exit code
    stderr : str
        Process stdout/stderr output
    profile : bytes
        If `profile` is `True` and run was at least partially successful,
        this key will be a byte string containing the cProfile data.
        Otherwise, None.
    """)


def skip_benchmarks(benchmarks, env, results=None):
    """
    Mark benchmarks as skipped.

    Parameters
    ----------
    benchmarks : Benchmarks
        Set of benchmarks to skip
    env : Environment
        Environment to skip them in
    results : Results, optional
        Where to store the results.
        If omitted, stored to a new unnamed Results object.

    Returns
    -------
    results : Results
        Benchmark results.

    """
    if results is None:
        results = Results.unnamed()

    log.warning("Skipping {0}".format(env.name))
    with log.indent():
        for name, benchmark in six.iteritems(benchmarks):
            log.step()
            log.warning('{0} skipped'.format(name))

            started_at = datetime.datetime.utcnow()
            r = fail_benchmark(benchmark)
            results.add_result(benchmark, r,
                               selected_idx=benchmarks.benchmark_selection.get(name),
                               started_at=started_at)

    return results


def run_benchmarks(benchmarks, env, results=None,
                   show_stderr=False, quick=False, profile=False,
                   extra_params=None,
                   record_samples=False, append_samples=False,
                   run_rounds=None,
                   launch_method=None):
    """
    Run all of the benchmarks in the given `Environment`.

    Parameters
    ----------
    benchmarks : Benchmarks
        Benchmarks to run
    env : Environment object
        Environment in which to run the benchmarks.
    results : Results, optional
        Where to store the results.
        If omitted, stored to a new unnamed Results object.
    show_stderr : bool, optional
        When `True`, display any stderr emitted by the benchmark.
    quick : bool, optional
        When `True`, run each benchmark function exactly once.
        This is useful to quickly find errors in the benchmark
        functions, without taking the time necessary to get
        accurate timings.
    profile : bool, optional
        When `True`, run the benchmark through the `cProfile`
        profiler.
    extra_params : dict, optional
        Override values for benchmark attributes.
    record_samples : bool, optional
        Whether to retain result samples or discard them.
    append_samples : bool, optional
        Whether to retain any previously measured result samples
        and use them in statistics computations.
    run_rounds : sequence of int, optional
        Run rounds for benchmarks with multiple rounds.
        If None, run all rounds.
    launch_method : {'auto', 'spawn', 'forkserver'}, optional
        Benchmark launching method to use.

    Returns
    -------
    results : Results
        Benchmark results.

    """

    if extra_params is None:
        extra_params = {}
    else:
        extra_params = dict(extra_params)

    if quick:
        extra_params['number'] = 1
        extra_params['repeat'] = 1
        extra_params['warmup_time'] = 0
        extra_params['rounds'] = 1

    if results is None:
        results = Results.unnamed()

    # Find all setup_cache routines needed
    setup_cache_timeout = {}
    benchmark_order = {}
    cache_users = {}
    max_rounds = 0

    def get_rounds(benchmark):
        """Get number of rounds to use for a job"""
        if 'rounds' in extra_params:
            return int(extra_params['rounds'])
        else:
            return int(benchmark.get('rounds', 1))

    for name, benchmark in sorted(six.iteritems(benchmarks)):
        key = benchmark.get('setup_cache_key')
        setup_cache_timeout[key] = max(benchmark.get('setup_cache_timeout',
                                                     benchmark['timeout']),
                                       setup_cache_timeout.get(key, 0))
        benchmark_order.setdefault(key, []).append((name, benchmark))
        max_rounds = max(max_rounds, get_rounds(benchmark))
        cache_users.setdefault(key, set()).add(name)

    if run_rounds is None:
        run_rounds = list(range(1, max_rounds + 1))

    # Interleave benchmark runs, in setup_cache order
    existing_results = results.get_result_keys(benchmarks)

    def iter_run_items():
        for run_round in run_rounds[::-1]:
            for setup_cache_key, benchmark_set in six.iteritems(benchmark_order):
                for name, benchmark in benchmark_set:
                    log.step()

                    rounds = get_rounds(benchmark)

                    if run_round > rounds:
                        if (not append_samples and
                                run_round == run_rounds[-1] and
                                name in existing_results):
                            # We need to remove samples here so that
                            # append_samples=False has an effect on all
                            # benchmarks regardless of whether they were
                            # run this round.
                            selected_idx = benchmarks.benchmark_selection.get(name)
                            results.remove_samples(name, selected_idx)
                        continue

                    is_final = (run_round == 1)
                    yield name, benchmark, setup_cache_key, is_final

    # Run benchmarks in order
    cache_dirs = {None: None}
    failed_benchmarks = set()
    failed_setup_cache = {}

    if append_samples:
        previous_result_keys = existing_results
    else:
        previous_result_keys = set()

    benchmark_durations = {}

    log.info("Benchmarking {0}".format(env.name))

    partial_info_time = None
    indent = log.indent()
    indent.__enter__()

    spawner = get_spawner(env, benchmarks.benchmark_dir,
                          launch_method=launch_method)

    try:
        # Preimport benchmark suite (if using forkserver)
        success, out = spawner.preimport()

        if success:
            if show_stderr and out:
                log.info("Importing benchmark suite produced output:")
                with log.indent():
                    log.error(out.rstrip())
        else:
            log.warning("Importing benchmark suite failed (skipping all benchmarks).")
            if show_stderr and out:
                with log.indent():
                    log.error(out)

            stderr = 'asv: benchmark suite import failed'
            for name, benchmark, setup_cache_key, is_final in iter_run_items():
                if name in failed_benchmarks:
                    continue

                selected_idx = benchmarks.benchmark_selection.get(name)
                started_at = datetime.datetime.utcnow()
                res = fail_benchmark(benchmark, stderr=stderr)
                results.add_result(benchmark, res,
                                   selected_idx=selected_idx,
                                   started_at=started_at,
                                   record_samples=record_samples)
                failed_benchmarks.add(name)
            return results

        # Run benchmarks
        for name, benchmark, setup_cache_key, is_final in iter_run_items():
            selected_idx = benchmarks.benchmark_selection.get(name)

            started_at = datetime.datetime.utcnow()

            # Don't try to rerun failed benchmarks
            if name in failed_benchmarks:
                if is_final:
                    partial_info_time = None
                    log.info(name, reserve_space=True)
                    log_benchmark_result(results, benchmark,
                                         show_stderr=show_stderr)
                continue

            # Setup cache first, if needed
            if setup_cache_key is None:
                cache_dir = None
            elif setup_cache_key in cache_dirs:
                cache_dir = cache_dirs[setup_cache_key]
            elif setup_cache_key not in failed_setup_cache:
                partial_info_time = None
                log.info("Setting up {0}".format(setup_cache_key), reserve_space=True)
                params_str = json.dumps({'cpu_affinity': extra_params.get('cpu_affinity')})
                cache_dir, stderr = spawner.create_setup_cache(
                    name, setup_cache_timeout[setup_cache_key], params_str)
                if cache_dir is not None:
                    log.add_padded('ok')
                    cache_dirs[setup_cache_key] = cache_dir
                else:
                    log.add_padded('failed')
                    if stderr and show_stderr:
                        with log.indent():
                            log.error(stderr)
                    failed_setup_cache[setup_cache_key] = stderr

                duration = (datetime.datetime.utcnow() - started_at).total_seconds()
                results.set_setup_cache_duration(setup_cache_key, duration)
                started_at = datetime.datetime.utcnow()

            if setup_cache_key in failed_setup_cache:
                # Mark benchmark as failed
                partial_info_time = None
                log.warning('{0} skipped (setup_cache failed)'.format(name))
                stderr = 'asv: setup_cache failed\n\n{}'.format(failed_setup_cache[setup_cache_key])
                res = fail_benchmark(benchmark, stderr=stderr)
                results.add_result(benchmark, res,
                                   selected_idx=selected_idx,
                                   started_at=started_at,
                                   record_samples=record_samples)
                failed_benchmarks.add(name)
                continue

            # If appending to previous results, make sure to use the
            # same value for 'number' attribute.
            cur_extra_params = extra_params
            if name in previous_result_keys:
                cur_extra_params = []
                prev_stats = results.get_result_stats(name, benchmark['params'])
                for s in prev_stats:
                    if s is None or 'number' not in s:
                        p = extra_params
                    else:
                        p = dict(extra_params)
                        p['number'] = s['number']
                    cur_extra_params.append(p)

            # Run benchmark
            if is_final:
                partial_info_time = None
                log.info(name, reserve_space=True)
            elif partial_info_time is None or time.time() > partial_info_time + 30:
                partial_info_time = time.time()
                log.info('Running ({0}--)'.format(name))

            res = run_benchmark(benchmark, spawner,
                                profile=profile,
                                selected_idx=selected_idx,
                                extra_params=cur_extra_params,
                                cwd=cache_dir)

            # Retain runtime durations
            ended_at = datetime.datetime.utcnow()
            if name in benchmark_durations:
                benchmark_durations[name] += (ended_at - started_at).total_seconds()
            else:
                benchmark_durations[name] = (ended_at - started_at).total_seconds()

            # Save result
            results.add_result(benchmark, res,
                               selected_idx=selected_idx,
                               started_at=started_at,
                               duration=benchmark_durations[name],
                               record_samples=(not is_final or record_samples),
                               append_samples=(name in previous_result_keys))

            previous_result_keys.add(name)

            if all(r is None for r in res.result):
                failed_benchmarks.add(name)

            # Log result
            if is_final:
                partial_info_time = None
                log_benchmark_result(results, benchmark,
                                     show_stderr=show_stderr)
            else:
                log.add('.')

            # Cleanup setup cache, if no users left
            if cache_dir is not None and is_final:
                cache_users[setup_cache_key].remove(name)
                if not cache_users[setup_cache_key]:
                    # No users of this cache left, perform cleanup
                    util.long_path_rmtree(cache_dir, True)
                    del cache_dirs[setup_cache_key]
    finally:
        # Cleanup any dangling caches
        for cache_dir in cache_dirs.values():
            if cache_dir is not None:
                util.long_path_rmtree(cache_dir, True)
        indent.__exit__(None, None, None)
        spawner.close()

    return results


def get_spawner(env, benchmark_dir, launch_method):
    has_fork = hasattr(os, 'fork') and hasattr(socket, 'AF_UNIX')

    if launch_method in (None, 'auto'):
        # Don't use ForkServer as default on OSX, because many Apple
        # things are not fork-safe
        if has_fork and sys.platform not in ('darwin',):
            launch_method = "forkserver"
        else:
            launch_method = "spawn"

    if launch_method == "spawn":
        spawner_cls = Spawner
    elif launch_method == "forkserver":
        if not has_fork:
            raise util.UserError("'forkserver' launch method not available "
                                 "on this platform")
        spawner_cls = ForkServer
    else:
        raise ValueError("Invalid launch_method: {}".format(launch_method))

    return spawner_cls(env, benchmark_dir)


def log_benchmark_result(results, benchmark, show_stderr=False):
    info, details = format_benchmark_result(results, benchmark)

    log.add_padded(info)
    if details:
        log.info(details, color='default')

    # Dump program output
    stderr = results.stderr.get(benchmark['name'])
    errcode = results.errcode.get(benchmark['name'])

    if errcode not in (None, 0, util.TIMEOUT_RETCODE, JSON_ERROR_RETCODE):
        # Display also error code
        if not stderr:
            stderr = ""
        else:
            stderr += "\n"
        stderr += "asv: benchmark failed (exit status {})".format(errcode)

    if stderr and show_stderr:
        with log.indent():
            log.error(stderr)


def fail_benchmark(benchmark, stderr='', errcode=1):
    """
    Return a BenchmarkResult describing a failed benchmark.
    """
    if benchmark['params']:
        # Mark only selected parameter combinations skipped
        params = itertools.product(*benchmark['params'])
        result = [None for idx in params]
        samples = [None] * len(result)
        number = [None] * len(result)
    else:
        result = [None]
        samples = [None]
        number = [None]

    return BenchmarkResult(result=result,
                           samples=samples,
                           number=number,
                           errcode=errcode,
                           stderr=stderr,
                           profile=None)


def run_benchmark(benchmark, spawner, profile,
                  selected_idx=None,
                  extra_params=None,
                  cwd=None,
                  prev_result=None):
    """
    Run a benchmark.

    Parameters
    ----------
    benchmark : dict
        Benchmark object dict
    spawner : Spawner
        Benchmark process spawner
    profile : bool
        Whether to run with profile
    selected_idx : set, optional
        Set of parameter indices to run for.
    extra_params : {dict, list}, optional
        Additional parameters to pass to the benchmark.
        If a list, each entry should correspond to a benchmark
        parameter combination.
    cwd : str, optional
        Working directory to run the benchmark in.
        If None, run in a temporary directory.

    Returns
    -------
    result : BenchmarkResult
        Result data.

    """

    if extra_params is None:
        extra_params = {}

    result = []
    samples = []
    number = []
    profiles = []
    stderr = ''
    errcode = 0

    if benchmark['params']:
        param_iter = enumerate(itertools.product(*benchmark['params']))
    else:
        param_iter = [(0, None)]

    for param_idx, params in param_iter:
        if selected_idx is not None and param_idx not in selected_idx:
            result.append(util.nan)
            samples.append(None)
            number.append(None)
            profiles.append(None)
            continue

        if isinstance(extra_params, list):
            cur_extra_params = extra_params[param_idx]
        else:
            cur_extra_params = extra_params

        res = _run_benchmark_single_param(
            benchmark, spawner, param_idx,
            extra_params=cur_extra_params, profile=profile,
            cwd=cwd)

        result += res.result
        samples += res.samples
        number += res.number

        profiles.append(res.profile)

        if res.stderr:
            stderr += "\n\n"
            stderr += res.stderr

        if res.errcode != 0:
            errcode = res.errcode

    return BenchmarkResult(
        result=result,
        samples=samples,
        number=number,
        errcode=errcode,
        stderr=stderr.strip(),
        profile=_combine_profile_data(profiles)
    )


def _run_benchmark_single_param(benchmark, spawner, param_idx,
                                profile, extra_params, cwd):
    """
    Run a benchmark, for single parameter combination index in case it
    is parameterized

    Parameters
    ----------
    benchmark : dict
        Benchmark object dict
    spawner : Spawner
        Benchmark process spawner
    param_idx : {int, None}
        Parameter index to run benchmark for
    profile : bool
        Whether to run with profile
    extra_params : dict
        Additional parameters to pass to the benchmark
    cwd : {str, None}
        Working directory to run the benchmark in.
        If None, run in a temporary directory.

    Returns
    -------
    result : BenchmarkResult
        Result data.

    """
    name = benchmark['name']
    if benchmark['params']:
        name += '-%d' % (param_idx,)

    if profile:
        profile_fd, profile_path = tempfile.mkstemp()
        os.close(profile_fd)
    else:
        profile_path = 'None'

    params_str = json.dumps(extra_params)

    if cwd is None:
        real_cwd = tempfile.mkdtemp()
    else:
        real_cwd = cwd

    result_file = tempfile.NamedTemporaryFile(delete=False)
    try:
        result_file.close()

        out, errcode = spawner.run(
            name=name, params_str=params_str, profile_path=profile_path,
            result_file_name=result_file.name,
            timeout=benchmark['timeout'],
            cwd=real_cwd)

        if errcode != 0:
            if errcode == util.TIMEOUT_RETCODE:
                out += "\n\nasv: benchmark timed out (timeout {0}s)\n".format(benchmark['timeout'])

            result = None
            samples = None
            number = None
        else:
            with open(result_file.name, 'r') as stream:
                data = stream.read()

            try:
                data = json.loads(data)
            except ValueError as exc:
                data = None
                errcode = JSON_ERROR_RETCODE
                out += "\n\nasv: failed to parse benchmark result: {0}\n".format(exc)

            # Special parsing for timing benchmark results
            if isinstance(data, dict) and 'samples' in data and 'number' in data:
                result = True
                samples = data['samples']
                number = data['number']
            else:
                result = data
                samples = None
                number = None

        if benchmark['params'] and out:
            params, = itertools.islice(itertools.product(*benchmark['params']),
                                       param_idx, param_idx + 1)
            out = "For parameters: {0}\n{1}".format(", ".join(params), out)

        if profile:
            with io.open(profile_path, 'rb') as profile_fd:
                profile_data = profile_fd.read()
            profile_data = profile_data if profile_data else None
        else:
            profile_data = None

        return BenchmarkResult(
            result=[result],
            samples=[samples],
            number=[number],
            errcode=errcode,
            stderr=out.strip(),
            profile=profile_data)

    except KeyboardInterrupt:
        spawner.interrupt()
        raise util.UserError("Interrupted.")
    finally:
        os.remove(result_file.name)
        if profile:
            os.remove(profile_path)
        if cwd is None:
            util.long_path_rmtree(real_cwd, True)


class Spawner(object):
    """
    Manage launching individual benchmark.py commands
    """

    def __init__(self, env, benchmark_dir):
        self.env = env
        self.benchmark_dir = os.path.abspath(benchmark_dir)
        self.interrupted = False

    def interrupt(self):
        self.interrupted = True

    def create_setup_cache(self, benchmark_id, timeout, params_str):
        cache_dir = tempfile.mkdtemp()

        env_vars = dict(os.environ)
        env_vars.update(self.env.env_vars)

        out, _, errcode = self.env.run(
            [BENCHMARK_RUN_SCRIPT, 'setup_cache',
             os.path.abspath(self.benchmark_dir),
             benchmark_id, params_str],
            dots=False, display_error=False,
            return_stderr=True, valid_return_codes=None,
            redirect_stderr=True,
            cwd=cache_dir,
            timeout=timeout,
            env=env_vars)

        if errcode == 0:
            return cache_dir, None
        else:
            util.long_path_rmtree(cache_dir, True)
            out += '\nasv: setup_cache failed (exit status {})'.format(errcode)
            return None, out.strip()

    def run(self, name, params_str, profile_path, result_file_name, timeout, cwd):
        env_vars = dict(os.environ)
        env_vars.update(self.env.env_vars)

        out, _, errcode = self.env.run(
            [BENCHMARK_RUN_SCRIPT, 'run', os.path.abspath(self.benchmark_dir),
             name, params_str, profile_path, result_file_name],
            dots=False, timeout=timeout,
            display_error=False, return_stderr=True, redirect_stderr=True,
            valid_return_codes=None, cwd=cwd,
            env=env_vars)
        return out, errcode

    def preimport(self):
        return True, ""

    def close(self):
        pass


class ForkServer(Spawner):
    def __init__(self, env, root):
        super(ForkServer, self).__init__(env, root)

        if not (hasattr(os, 'fork') and hasattr(os, 'setpgid')):
            raise RuntimeError("ForkServer only available on POSIX")

        self.tmp_dir = tempfile.mkdtemp(prefix='asv-forkserver-')
        self.socket_name = os.path.join(self.tmp_dir, 'socket')

        env_vars = dict(os.environ)
        env_vars.update(env.env_vars)

        self.server_proc = env.run(
            [BENCHMARK_RUN_SCRIPT, 'run_server', self.benchmark_dir, self.socket_name],
            return_popen=True,
            redirect_stderr=True,
            env=env_vars)

        self._server_output = None
        self.stdout_reader_thread = threading.Thread(target=self._stdout_reader)
        self.stdout_reader_thread.start()

        # Wait for the socket to appear
        while self.stdout_reader_thread.is_alive():
            if os.path.exists(self.socket_name):
                break
            time.sleep(0.05)

        if not os.path.exists(self.socket_name):
            os.rmdir(self.tmp_dir)
            raise RuntimeError("Failed to start server thread")

    def _stdout_reader(self):
        try:
            out = self.server_proc.stdout.read()
            self.server_proc.stdout.close()
            out = out.decode('utf-8', 'replace')
        except Exception as exc:
            import traceback
            out = traceback.format_exc()

        self._server_output = out

    def run(self, name, params_str, profile_path, result_file_name, timeout, cwd):
        msg = {'action': 'run',
               'benchmark_id': name,
               'params_str': params_str,
               'profile_path': profile_path,
               'result_file': result_file_name,
               'timeout': timeout,
               'cwd': cwd}
        result = self._send_command(msg)
        return result['out'], result['errcode']

    def preimport(self):
        success = True
        out = ""
        try:
            out = self._send_command({'action': 'preimport'})
        except Exception as exc:
            success = False
            out = "asv: benchmark runner crashed\n"
            if isinstance(exc, util.UserError):
                out += str(exc)
            else:
                out += traceback.format_exc()
            out = out.rstrip()

        return success, out

    def _send_command(self, msg):
        msg = json.dumps(msg)
        if sys.version_info[0] >= 3:
            msg = msg.encode('utf-8')

        # Connect (with wait+retry)
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        for retry in range(5, 0, -1):
            try:
                s.connect(self.socket_name)
                break
            except socket.error:
                if retry > 1:
                    time.sleep(0.2)
                else:
                    raise

        # Send command
        try:
            s.sendall(struct.pack('<Q', len(msg)))
            s.sendall(msg)

            # Read result
            read_size, = struct.unpack('<Q', util.recvall(s, 8))
            result_text = util.recvall(s, read_size)
            if sys.version_info[0] >= 3:
                result_text = result_text.decode('utf-8')
            result = json.loads(result_text)
        except Exception:
            exitcode = self.server_proc.poll()
            if exitcode is not None:
                raise util.UserError("Process exited with code {0}".format(exitcode))
            raise
        finally:
            s.close()

        return result

    def close(self):
        import signal

        # Check for termination
        if self.server_proc.poll() is None:
            util._killpg_safe(self.server_proc.pid, signal.SIGINT)

        if self.server_proc.poll() is None:
            time.sleep(0.1)

        if self.server_proc.poll() is None:
            # Kill process group
            util._killpg_safe(self.server_proc.pid, signal.SIGKILL)

        self.server_proc.wait()
        self.stdout_reader_thread.join()

        if self._server_output and not self.interrupted:
            with log.indent():
                log.error("asv: forkserver:")
                log.error(self._server_output)

        util.long_path_rmtree(self.tmp_dir)


def _combine_profile_data(datasets):
    """
    Combine a list of profile data to a single profile
    """
    datasets = [data for data in datasets if data is not None]
    if not datasets:
        return None
    elif len(datasets) == 1:
        return datasets[0]

    # Load and combine stats
    stats = None

    while datasets:
        data = datasets.pop(0)

        f = tempfile.NamedTemporaryFile(delete=False)
        try:
            f.write(data)
            f.close()
            if stats is None:
                stats = pstats.Stats(f.name)
            else:
                stats.add(f.name)
        finally:
            os.remove(f.name)

    # Write combined stats out
    f = tempfile.NamedTemporaryFile(delete=False)
    try:
        f.close()
        stats.dump_stats(f.name)
        with open(f.name, 'rb') as fp:
            return fp.read()
    finally:
        os.remove(f.name)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
from collections import defaultdict

from . import Command
from ..benchmarks import Benchmarks
from ..machine import iter_machine_files
from ..results import iter_results_for_machine, iter_results_for_machine_and_hash
from ..runner import format_benchmark_result
from ..repo import get_repo, NoSuchNameError
from ..util import load_json
from ..console import log, color_print
from ..environment import get_environments
from .. import util

from . import common_args


class Show(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "show", help="Print recorded data",
            description="""Print saved benchmark results.""")

        parser.add_argument(
            'commit', nargs='?', default=None,
            help="""The commit to show data for""")
        parser.add_argument(
            '--details', action='store_true', default=False,
            help="""Show all result details""")
        parser.add_argument(
            '--durations', action='store_true', default=False,
            help="""Show only run durations""")
        common_args.add_bench(parser)
        common_args.add_machine(parser)
        common_args.add_environment(parser)
        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args, **kwargs):
        return cls.run(
            conf=conf, commit=args.commit, bench=args.bench,
            machine=args.machine, env_spec=args.env_spec,
            details=args.details, durations=args.durations,
            **kwargs
        )

    @classmethod
    def run(cls, conf, commit=None, bench=None, machine=None, env_spec=None,
            details=False, durations=False):
        if env_spec:
            env_names = ([env.name for env in get_environments(conf, env_spec, verbose=False)]
                         + list(env_spec))
        else:
            env_names = None

        machines = []
        for path in iter_machine_files(conf.results_dir):
            d = load_json(path)
            machines.append(d['machine'])

        if len(machines) == 0:
            raise util.UserError("No results found")
        elif machine is None:
            pass
        elif machine in machines:
            machines = [machine]
        else:
            raise util.UserError(
                "Results for machine '{0} not found".format(machine))

        benchmarks = Benchmarks.load(conf, regex=bench)

        if commit is None:
            result_iter = cls._iter_results(conf, machines, env_names)
            if durations:
                cls._print_commit_durations(conf, result_iter, benchmarks)
            else:
                cls._print_commits(conf, result_iter, benchmarks)
        else:
            result_iter = cls._iter_results(conf, machines, env_names, commit)
            if durations:
                cls._print_result_durations(conf, commit, result_iter, benchmarks)
            else:
                cls._print_results(conf, commit, result_iter,
                                   benchmarks, show_details=details)

    @classmethod
    def _iter_results(cls, conf, machines, env_names, commit_hash=None):
        """
        Iterate over results for given machines/environments.

        Yields
        ------
        machine : str
            Machine name
        result : asv.result.Results
            Results
        """
        if commit_hash is not None:
            repo = get_repo(conf)
            try:
                commit_hash = repo.get_hash_from_name(commit_hash)
            except NoSuchNameError:
                pass

        for machine in machines:
            if commit_hash is not None:
                result_iter = iter_results_for_machine_and_hash(
                    conf.results_dir, machine, commit_hash)
            else:
                result_iter = iter_results_for_machine(
                    conf.results_dir, machine)

            for result in result_iter:
                if env_names is not None and result.env_name not in env_names:
                    continue

                yield machine, result

    @classmethod
    def _print_commits(cls, conf, result_iter, benchmarks):
        commits = defaultdict(lambda: {})

        for machine, result in result_iter:
            if result.get_result_keys(benchmarks):
                commits[(machine, result.env_name)][result.commit_hash] = result.date

        log.flush()

        color_print("Commits with results:")
        color_print("")

        for machine, env_name in sorted(commits.keys()):
            color_print("Machine    : {}".format(machine))
            color_print("Environment: {}".format(env_name))
            color_print("")

            cur_commits = commits[(machine, env_name)]
            commit_order = list(cur_commits.keys())
            commit_order.sort(key=lambda x: cur_commits[x])

            for commit in commit_order:
                color_print("    {}".format(commit[:conf.hash_length]))

            color_print("")

    @classmethod
    def _print_results(cls, conf, commit_hash, result_iter, benchmarks, show_details=False):
        repo = get_repo(conf)

        log.flush()

        color_print("Commit: {}".format(repo.get_decorated_hash(commit_hash,
                                                                conf.hash_length)),
                    "blue")
        color_print("")

        for machine, result in result_iter:
            for name in sorted(result.get_result_keys(benchmarks)):
                cls._print_benchmark(machine, result, benchmarks[name],
                                     show_details=show_details)

    @classmethod
    def _print_benchmark(cls, machine, result, benchmark, show_details=False):
        color_print("{} [{}/{}]".format(benchmark['name'],
                                        machine,
                                        result.env_name),
                    'green')

        info, details = format_benchmark_result(result, benchmark)
        color_print("  {}".format(info), 'red')
        if details:
            color_print("  " + details.replace("\n", "\n  "))

        started_at = result.started_at.get(benchmark['name'])
        if started_at is not None:
            started_at = util.js_timestamp_to_datetime(started_at)
            started_at = started_at.strftime('%Y-%m-%d %H:%M:%S')
        else:
            started_at = "n/a"

        duration = result.duration.get(benchmark['name'])
        if duration is not None:
            duration = util.human_time(duration)
        else:
            duration = "n/a"

        if started_at != "n/a" or duration != "n/a":
            color_print('  started: {}, duration: {}'.format(
                started_at, duration))

        if not show_details:
            color_print("")
            return

        stats = result.get_result_stats(benchmark['name'], benchmark['params'])

        def get_stat_info(key):
            if key == 'ci_99':
                return [(x.get('ci_99_a'), x.get('ci_99_b')) if x is not None else None
                        for x in stats]
            return [x.get(key) if x is not None else None for x in stats]

        for key in ['repeat', 'number', 'ci_99', 'mean', 'std', 'min', 'max']:
            values = get_stat_info(key)

            if key == 'ci_99':
                values = ["({}, {})".format(util.human_value(x[0], benchmark['unit']),
                                            util.human_value(x[1], benchmark['unit']))
                          if x is not None else None
                          for x in values]
            elif any(isinstance(x, float) for x in values):
                values = [util.human_value(x, benchmark['unit']) if x is not None else None
                          for x in values]

            if not all(x is None for x in values):
                color_print("  {}: {}".format(key, ", ".join(map(str, values))))

        samples = result.get_result_samples(benchmark['name'], benchmark['params'])
        if not all(x is None for x in samples):
            color_print("  samples: {}".format(samples))

        color_print("")

    @classmethod
    def _get_durations(cls, result_iter, benchmarks, commits=False):
        durations = defaultdict(lambda: {})
        for machine, result in result_iter:
            total_duration = None

            keys = list(result.get_result_keys(benchmarks))
            keys += ["<build>"]

            for key in result.get_result_keys(benchmarks):
                setup_cache_key = benchmarks[key].get('setup_cache_key')
                if setup_cache_key is not None:
                    keys.append("<setup_cache {}>".format(setup_cache_key))

            for key in keys:
                duration = result.duration.get(key)
                if duration is not None:
                    if total_duration is None:
                        total_duration = 0
                    total_duration += duration

                    if not commits:
                        durations[(machine, result.env_name)][key] = duration

            if total_duration is None:
                total_duration = float("nan")

            if commits:
                durations[(machine, result.env_name)][result.commit_hash] = (result.date, total_duration)

        return durations

    @classmethod
    def _print_commit_durations(cls, conf, result_iter, benchmarks):
        durations = cls._get_durations(result_iter, benchmarks, commits=True)

        log.flush()

        color_print("Run durations:")
        color_print("")

        for machine, env_name in sorted(durations.keys()):
            color_print("Machine    : {}".format(machine))
            color_print("Environment: {}".format(env_name))
            color_print("")

            cur_durations = durations[(machine, env_name)]
            commit_order = list(cur_durations.keys())
            commit_order.sort(key=lambda x: cur_durations[x])

            for commit in commit_order:
                seconds = cur_durations[commit][1]
                color_print("    {}  {}".format(
                    commit, util.human_time(seconds)))

            color_print("")

    @classmethod
    def _print_result_durations(cls, conf, commit_hash, result_iter, benchmarks):
        durations = cls._get_durations(result_iter, benchmarks, commits=False)

        repo = get_repo(conf)

        log.flush()

        color_print("Commit: {}".format(repo.get_decorated_hash(commit_hash,
                                                                conf.hash_length)),
                    "blue")
        color_print("")

        for machine, env_name in sorted(durations.keys()):
            color_print("Machine    : {}".format(machine))
            color_print("Environment: {}".format(env_name))
            color_print("")

            cur_durations = durations[(machine, env_name)]
            order = list(cur_durations.keys())
            order.sort(key=lambda x: -cur_durations[x])

            total = 0

            for name in order:
                seconds = cur_durations[name]
                total += seconds
                color_print("    {}  {}".format(
                    name, util.human_time(seconds)))

            color_print("")
            color_print("    total duration: {}".format(util.human_time(total)))



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from .main import main

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import six

import os
import multiprocessing as mp
import platform
import sys
import textwrap

from . import console
from .console import log, color_print
from . import util


def iter_machine_files(results_dir):
    """
    Iterate over all of the machine.json files in the results_dir
    """
    for root, dirs, files in os.walk(results_dir):
        for filename in files:
            if filename == 'machine.json':
                path = os.path.join(root, filename)
                yield path


def _get_unique_machine_name():
    (system, node, release, version, machine, processor) = platform.uname()
    return node


class MachineCollection(object):
    """
    Stores information about 1 or more machines in the
    ~/.asv-machine.json file.
    """
    api_version = 1

    @staticmethod
    def get_machine_file_path():
        return os.path.expanduser('~/.asv-machine.json')

    @classmethod
    def load(cls, machine_name, _path=None):
        if _path is None:
            path = cls.get_machine_file_path()
        else:
            path = _path

        d = {}
        if os.path.isfile(path):
            d = util.load_json(path, cls.api_version)
            if machine_name in d:
                return d[machine_name]
            elif len(d) == 1 and machine_name == _get_unique_machine_name():
                return d[list(d.keys())[0]]

        raise util.UserError(
            "No information stored about machine '{0}'. I know about {1}.".format(
                machine_name, util.human_list(d.keys())))

    @classmethod
    def save(cls, machine_name, machine_info, _path=None):
        if _path is None:
            path = cls.get_machine_file_path()
        else:
            path = _path
        if os.path.isfile(path):
            d = util.load_json(path)
        else:
            d = {}
        d[machine_name] = machine_info
        util.write_json(path, d, cls.api_version)

    @classmethod
    def update(cls, _path=None):
        if _path is None:
            path = cls.get_machine_file_path()
        else:
            path = _path
        if os.path.isfile(path):
            util.update_json(cls, path, cls.api_version)


class Machine(object):
    """
    Stores information about a particular machine.
    """
    api_version = 1

    fields = [
        ("machine",
         """
         A unique name to identify this machine in the results.  May
         be anything, as long as it is unique across all the machines used
         to benchmark this project.

         NOTE: If changed from the default, it will no longer match
         the hostname of this machine, and you may need to explicitly
         use the --machine argument to asv.
         """),

        ("os",
         """
         The OS type and version of this machine.  For example,
         'Macintosh OS-X 10.8'."""),

        ("arch",
         """
         The generic CPU architecture of this machine.  For
         example, 'i386' or 'x86_64'."""),

        ("cpu",
         """
         A specific description of the CPU of this machine,
         including its speed and class.  For example, 'Intel(R)
         Core(TM) i5-2520M CPU @ 2.50GHz (4 cores)'."""),

        ("num_cpu",
         """
         The number of CPUs in the system. For example,
         '4'."""),

        ("ram",
         """
         The amount of physical RAM on this machine.  For example,
         '4GB'.""")
    ]

    hardcoded_machine_name = None

    @classmethod
    def get_unique_machine_name(cls):
        if cls.hardcoded_machine_name:
            return cls.hardcoded_machine_name
        return _get_unique_machine_name()

    @staticmethod
    def get_defaults():
        (system, node, release, version, machine, processor) = platform.uname()

        cpu = util.get_cpu_info()

        ram = six.text_type(util.get_memsize())

        num_cpu = six.text_type(mp.cpu_count())

        return {
            'machine': node,
            'os': "{0} {1}".format(system, release),
            'num_cpu': num_cpu,
            'arch': platform.machine(),
            'cpu': cpu,
            'ram': ram
            }

    @staticmethod
    def generate_machine_file(use_defaults=False):
        if not sys.stdout.isatty() and not use_defaults:
            raise util.UserError(
                "Run asv at the console the first time to generate "
                "one, or run `asv machine --yes`.")

        log.flush()

        color_print(
            "I will now ask you some questions about this machine to "
            "identify it in the benchmarks.")
        color_print("")

        defaults = Machine.get_defaults()
        values = {}

        for i, (name, description) in enumerate(Machine.fields):
            print(
                textwrap.fill(
                    '{0}. {1}: {2}'.format(
                        i+1, name, textwrap.dedent(description)),
                    subsequent_indent='   '))
            values[name] = console.get_answer_default(name, defaults[name],
                                                      use_defaults=use_defaults)

        return values

    @classmethod
    def load(cls, interactive=False, force_interactive=False, _path=None,
             machine_name=None, use_defaults=False, **kwargs):
        self = Machine()

        if machine_name is None:
            machine_name = cls.get_unique_machine_name()
        try:
            d = MachineCollection.load(machine_name, _path=_path)
        except util.UserError as e:
            console.log.error(str(e) + '\n')
            d = {}
        d.update(kwargs)
        if (not len(d) and interactive) or force_interactive:
            d.update(self.generate_machine_file(use_defaults=use_defaults))

        machine_name = d['machine']

        self.__dict__.update(d)
        MachineCollection.save(machine_name, self.__dict__, _path=_path)
        return self

    def save(self, results_dir):
        path = os.path.join(results_dir, self.machine, 'machine.json')
        util.write_json(path, self.__dict__, self.api_version)

    @classmethod
    def update(cls, path):
        util.update_json(cls, path, cls.api_version)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
A set of utilities for writing output to the console.
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import io
import contextlib
import locale
import logging
import os
import sys
import textwrap
import time

import six
from six.moves import xrange, input

from . import util


WIN = (os.name == "nt")


def isatty(file):
    """
    Returns `True` if `file` is a tty.

    Most built-in Python file-like objects have an `isatty` member,
    but some user-defined types may not, so this assumes those are not
    ttys.
    """
    if hasattr(file, 'isatty'):
        return file.isatty()
    return False


def _color_text(text, color):
    """
    Returns a string wrapped in ANSI color codes for coloring the
    text in a terminal::

        colored_text = color_text('Here is a message', 'blue')

    This won't actually effect the text until it is printed to the
    terminal.

    Parameters
    ----------
    text : str
        The string to return, bounded by the color codes.
    color : str
        An ANSI terminal color name. Must be one of:
        black, red, green, brown, blue, magenta, cyan, lightgrey,
        default, darkgrey, lightred, lightgreen, yellow, lightblue,
        lightmagenta, lightcyan, white, or '' (the empty string).
    """
    color_mapping = {
        'black': '0;30',
        'red': '0;31',
        'green': '0;32',
        'brown': '0;33',
        'blue': '0;34',
        'magenta': '0;35',
        'cyan': '0;36',
        'lightgrey': '0;37',
        'default': '0;39',
        'darkgrey': '1;30',
        'lightred': '1;31',
        'lightgreen': '1;32',
        'yellow': '1;33',
        'lightblue': '1;34',
        'lightmagenta': '1;35',
        'lightcyan': '1;36',
        'white': '1;37'}

    color_code = color_mapping.get(color, '0;39')
    return '\033[{0}m{1}\033[0m'.format(color_code, text)


# This is a table of Unicode characters that we want to have
# reasonable representations in ascii so they aren't just replaced
# with '?'.  A complete solution to this problem would involve a
# third-party library such as "unidecode", but this handles the common
# cases of stuff coming from asv.
#
# You can find the characters that need an entry using:
#    grep -P  -n '[^\x00-\x7F]' -r *
# in the `asv` source directory.

_unicode_translations = {
    ord('μ'): 'u',
    ord('·'): '-',
    ord('±'): '~'
}


def _write_with_fallback(s, fileobj):
    """
    Write the supplied string to the given stream, but switch to
    the locale's preferred encoding in case of a UnicodeEncodeError.
    Failing that, replace characters.
    *fileobj* must be text stream on Py3, on Py2 a `file` byte stream.
    *s* must be unicode.
    """
    if not isinstance(s, six.text_type):
        raise ValueError("Input string is not a Unicode string")

    if six.PY3:
        try:
            fileobj.write(s)
            return
        except UnicodeError:
            pass

        # Fall back to writing bytes
        enc = locale.getpreferredencoding()
        try:
            b = s.encode(enc)
        except UnicodeError:
            s = s.translate(_unicode_translations)
            b = s.encode(enc, errors='replace')

        fileobj.flush()
        fileobj.buffer.write(b)
    else:
        enc = locale.getpreferredencoding()
        try:
            b = s.encode(enc)
        except UnicodeError:
            for key, val in _unicode_translations.iteritems():
                s = s.replace(unichr(key), val)
            b = s.encode(enc, errors='replace')

        fileobj.write(b)


def color_print(*args, **kwargs):
    """
    Prints colors and styles to the terminal uses ANSI escape
    sequences.

    ::

       color_print('This is the color ', 'default', 'GREEN', 'green')

    Parameters
    ----------
    positional args : str
        The positional arguments come in pairs (*msg*, *color*), where
        *msg* is the string to display and *color* is the color to
        display it in.

        *color* is an ANSI terminal color name.  Must be one of:
        black, red, green, brown, blue, magenta, cyan, lightgrey,
        default, darkgrey, lightred, lightgreen, yellow, lightblue,
        lightmagenta, lightcyan, white, or '' (the empty string).

    file : writeable file-like object, optional
        Where to write to.  Defaults to `sys.stdout`.  If file is not
        a tty (as determined by calling its `isatty` member, if one
        exists), no coloring will be included.

    end : str, optional
        The ending of the message.  Defaults to ``\\n``.  The end will
        be printed after resetting any color or font state.
    """

    file = kwargs.get('file', sys.stdout)
    end = kwargs.get('end', '\n')

    if isatty(file) and not WIN:
        for i in xrange(0, len(args), 2):
            msg = args[i]
            if i + 1 == len(args):
                color = ''
            else:
                color = args[i + 1]

            if color:
                msg = _color_text(msg, color)
            _write_with_fallback(msg, file)

        _write_with_fallback(end, file)
    else:
        for i in xrange(0, len(args), 2):
            msg = args[i]
            _write_with_fallback(msg, file)
        _write_with_fallback(end, file)


def get_answer_default(prompt, default, use_defaults=False):
    color_print("{0} [{1}]: ".format(prompt, default), end='')

    if use_defaults:
        return default

    x = input()
    if x.strip() == '':
        return default
    return x


def truncate_left(s, l):
    if len(s) > l:
        return '...' + s[-(l - 3):]
    else:
        return s


class Log(object):
    def __init__(self):
        self._indent = 1
        self._total = 0
        self._count = 0
        self._logger = logging.getLogger()
        self._needs_newline = False
        self._last_dot = time.time()

    def _stream_formatter(self, record):
        '''
        The formatter for standard output
        '''
        if self._needs_newline:
            color_print('')
        parts = record.msg.split('\n', 1)
        first_line = parts[0]
        if len(parts) == 1:
            rest = None
        else:
            rest = parts[1]

        indent = self._indent + 1
        continued = getattr(record, 'continued', False)

        if self._total:
            progress_msg = '[{0:6.02f}%] '.format(
                (float(self._count) / self._total) * 100.0)
            if not continued:
                color_print(progress_msg, end='')
            indent += len(progress_msg)

        if not continued:
            color_print('·' * self._indent, end='')
            color_print(' ', end='')
        else:
            color_print(' ' * indent, end='')

        if hasattr(record, 'color'):
            color = record.color
        elif record.levelno < logging.DEBUG:
            color = 'default'
        elif record.levelno < logging.INFO:
            color = 'default'
        elif record.levelno < logging.WARN:
            if self._indent == 1:
                color = 'green'
            elif self._indent == 2:
                color = 'blue'
            else:
                color = 'default'
        elif record.levelno < logging.ERROR:
            color = 'brown'
        else:
            color = 'red'

        spaces = ' ' * indent
        color_print(first_line, color, end='')
        if rest is not None:
            color_print('')
            detail = textwrap.dedent(rest)
            for line in detail.split('\n'):
                color_print(spaces, end='')
                color_print(line)

        self._needs_newline = True
        sys.stdout.flush()

    @contextlib.contextmanager
    def indent(self):
        """
        A context manager to increase the indentation level.
        """
        self._indent += 1
        yield
        self._indent -= 1

    def dot(self):
        if isatty(sys.stdout):
            if time.time() > self._last_dot + 1.0:
                color_print('.', 'darkgrey', end='')
                sys.stdout.flush()
                self._last_dot = time.time()

    def set_nitems(self, n):
        """
        Set the number of remaining items to process.  Each of these
        steps should be incremented through using `step`.

        Can be called multiple times. The progress percentage is ensured
        to be non-decreasing, except if 100% was already reached in which
        case it is restarted from 0%.
        """
        try:
            # Ensure count/total is nondecreasing
            self._total = util.ceildiv(n * self._total, self._total - self._count)
            self._count = self._total - n
        except ZeroDivisionError:
            # Reset counting from start
            self._total = n
            self._count = 0

    def step(self):
        """
        Write that a step has been completed.  A percentage is
        displayed along with it.

        If we are stepping beyond the number of items, stop counting.
        """
        self._count = min(self._total, self._count + 1)

    def enable(self, verbose=False):
        sh = logging.StreamHandler()
        sh.emit = self._stream_formatter
        self._logger.addHandler(sh)
        if verbose:
            self._logger.setLevel(logging.DEBUG)
        else:
            self._logger.setLevel(logging.INFO)

    @contextlib.contextmanager
    def set_level(self, level):
        orig_level = self._logger.level
        if not self.is_debug_enabled():
            self._logger.setLevel(level)
        try:
            yield
        finally:
            self._logger.setLevel(orig_level)

    def is_debug_enabled(self):
        return self._logger.getEffectiveLevel() <= logging.DEBUG

    def _message(self, routine, message, reserve_space=False, color=None,
                 continued=False):
        kwargs = {}
        extra = {}
        if color is not None:
            extra['color'] = color
        if continued:
            extra['continued'] = True
        if extra:
            kwargs['extra'] = extra

        if reserve_space:
            max_width = max(16, util.get_terminal_width() - 33)
            message = truncate_left(message, max_width)
            self._prev_message = message

        routine(message, **kwargs)

    def info(self, *args, **kwargs):
        self._message(self._logger.info, *args, **kwargs)

    def warning(self, *args, **kwargs):
        self._message(self._logger.warning, *args, **kwargs)

    def debug(self, *args, **kwargs):
        self._message(self._logger.debug, *args, **kwargs)

    def error(self, *args, **kwargs):
        self._message(self._logger.error, *args, **kwargs)

    def add(self, msg):
        if self._needs_newline:
            _write_with_fallback(msg, sys.stdout)
            sys.stdout.flush()
        else:
            self.info(msg)

    def add_padded(self, msg):
        """
        Final part of two-part info message.
        Should be preceded by a call to info/warn/...(msg, reserve_space=True)
        """
        if self._prev_message is None:
            # No previous part: print as an info message
            self.info(msg)
            return

        padding_length = util.get_terminal_width() - len(self._prev_message) - 14 - 1 - len(msg)
        if WIN:
            padding_length -= 1
        padding = " "*padding_length

        self._prev_message = None
        self.add(" {0}{1}".format(padding, msg))

    def flush(self):
        """
        Flush any trailing newlines. Needs to be called before printing
        to stdout via other means, after using Log.
        """
        if self._needs_newline:
            color_print('')
            self._needs_newline = False
        sys.stdout.flush()


log = Log()



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""\
Usage: python -masv.benchmark COMMAND [...]

Manage a single benchmark and, when run from the commandline, report
its runtime to a file.

commands:

  timing [...]
      Run timing benchmark for given Python statement.

internal commands:

  discover BENCHMARK_DIR RESULT_FILE
      Discover benchmarks in a given directory and store result to a file.
  setup_cache BENCHMARK_DIR BENCHMARK_ID
      Run setup_cache for given benchmark.
  run BENCHMARK_DIR BENCHMARK_ID QUICK PROFILE_PATH RESULT_FILE
      Run a given benchmark, and store result in a file.
  run_server BENCHMARK_DIR SOCKET_FILENAME
      Run a Unix socket forkserver.
"""

# !!!!!!!!!!!!!!!!!!!! NOTE !!!!!!!!!!!!!!!!!!!!
# This file, unlike most others, must be compatible with as many
# versions of Python as possible and have no dependencies outside of
# the Python standard library.  This is the only bit of code from asv
# that is imported into the benchmarking process.

# Remove asv package directory from sys.path. This script file resides
# there although it's not part of the package, and Python puts it to
# sys.path[0] on start which can shadow other modules
import sys
if __name__ == "__main__":
    _old_sys_path_head = sys.path.pop(0)
else:
    _old_sys_path_head = None

import copy
try:
    import cProfile as profile
except:
    profile = None
import ctypes
from ctypes.util import find_library
from hashlib import sha256
import errno
if sys.version_info[0] >= 3:
    import importlib.machinery
else:
    import imp
import inspect
import itertools
import json
import os
import pickle
import re
import subprocess
import textwrap
import timeit
import time
import tempfile
import struct
import pkgutil
import traceback
import contextlib
from importlib import import_module
from collections import Counter
import warnings


# The best timer we can use is time.process_time, but it is not
# available in the Python stdlib until Python 3.3.  This is a ctypes
# backport for Pythons that don't have it.

try:
    from time import process_time
except ImportError:  # Python <3.3
    if sys.platform.startswith("linux"):
        CLOCK_PROCESS_CPUTIME_ID = 2  # time.h

        clockid_t = ctypes.c_int
        time_t = ctypes.c_long

        class timespec(ctypes.Structure):
            _fields_ = [
                ('tv_sec', time_t),         # seconds
                ('tv_nsec', ctypes.c_long)  # nanoseconds
            ]
        _clock_gettime = ctypes.CDLL(
            find_library('rt'), use_errno=True).clock_gettime
        _clock_gettime.argtypes = [clockid_t, ctypes.POINTER(timespec)]

        def process_time():
            tp = timespec()
            if _clock_gettime(CLOCK_PROCESS_CPUTIME_ID, ctypes.byref(tp)) < 0:
                err = ctypes.get_errno()
                msg = errno.errorcode[err]
                if err == errno.EINVAL:
                    msg += (
                        "The clk_id (4) specified is not supported on this system")
                raise OSError(err, msg)
            return tp.tv_sec + tp.tv_nsec * 1e-9

    elif sys.platform == 'darwin':
        RUSAGE_SELF = 0  # sys/resources.h

        time_t = ctypes.c_long
        suseconds_t = ctypes.c_int32

        class timeval(ctypes.Structure):
            _fields_ = [
                ('tv_sec', time_t),
                ('tv_usec', suseconds_t)
            ]

        class rusage(ctypes.Structure):
            _fields_ = [
                ('ru_utime', timeval),
                ('ru_stime', timeval),
                ('ru_maxrss', ctypes.c_long),
                ('ru_ixrss', ctypes.c_long),
                ('ru_idrss', ctypes.c_long),
                ('ru_isrss', ctypes.c_long),
                ('ru_minflt', ctypes.c_long),
                ('ru_majflt', ctypes.c_long),
                ('ru_nswap', ctypes.c_long),
                ('ru_inblock', ctypes.c_long),
                ('ru_oublock', ctypes.c_long),
                ('ru_msgsnd', ctypes.c_long),
                ('ru_msgrcv', ctypes.c_long),
                ('ru_nsignals', ctypes.c_long),
                ('ru_nvcsw', ctypes.c_long),
                ('ru_nivcsw', ctypes.c_long)
            ]

        _getrusage = ctypes.CDLL(find_library('c'), use_errno=True).getrusage
        _getrusage.argtypes = [ctypes.c_int, ctypes.POINTER(rusage)]

        def process_time():
            ru = rusage()
            if _getrusage(RUSAGE_SELF, ctypes.byref(ru)) < 0:
                err = ctypes.get_errno()
                msg = errno.errorcode[err]
                if err == errno.EINVAL:
                    msg += (
                        "The clk_id (0) specified is not supported on this system")
                raise OSError(err, msg)
            return float(ru.ru_utime.tv_sec + ru.ru_utime.tv_usec * 1e-6 +
                         ru.ru_stime.tv_sec + ru.ru_stime.tv_usec * 1e-6)

    else:
        # Fallback to default timer
        process_time = timeit.default_timer

wall_timer = timeit.default_timer


def get_maxrss():
    # Fallback function, in case we don't have one that works on the
    # current platform
    return None

if sys.platform.startswith('win'):
    import ctypes
    import ctypes.wintypes

    SIZE_T = ctypes.c_size_t
    class PROCESS_MEMORY_COUNTERS(ctypes.Structure):
        _fields_ = [
            ('cb', ctypes.wintypes.DWORD),
            ('PageFaultCount', ctypes.wintypes.DWORD),
            ('PeakWorkingSetSize', SIZE_T),
            ('WorkingSetSize', SIZE_T),
            ('QuotaPeakPagedPoolUsage', SIZE_T),
            ('QuotaPagedPoolUsage', SIZE_T),
            ('QuotaPeakNonPagedPoolUsage', SIZE_T),
            ('QuotaNonPagedPoolUsage', SIZE_T),
            ('PagefileUsage', SIZE_T),
            ('PeakPagefileUsage', SIZE_T),
        ]

    GetCurrentProcess = ctypes.windll.kernel32.GetCurrentProcess
    GetCurrentProcess.argtypes = []
    GetCurrentProcess.restype = ctypes.wintypes.HANDLE

    GetProcessMemoryInfo = ctypes.windll.psapi.GetProcessMemoryInfo
    GetProcessMemoryInfo.argtypes = (ctypes.wintypes.HANDLE,
                                     ctypes.POINTER(PROCESS_MEMORY_COUNTERS),
                                     ctypes.wintypes.DWORD)
    GetProcessMemoryInfo.restype = ctypes.wintypes.BOOL

    def get_maxrss():
        proc_hnd = GetCurrentProcess()
        counters = PROCESS_MEMORY_COUNTERS()
        info = GetProcessMemoryInfo(proc_hnd, ctypes.byref(counters), ctypes.sizeof(counters))
        if info == 0:
            raise ctypes.WinError()
        return counters.PeakWorkingSetSize

    if ctypes.sizeof(ctypes.c_void_p) == ctypes.sizeof(ctypes.c_uint64):
        DWORD_PTR = ctypes.c_uint64
    elif ctypes.sizeof(ctypes.c_void_p) == ctypes.sizeof(ctypes.c_uint32):
        DWORD_PTR = ctypes.c_uint32

    SetProcessAffinityMask = ctypes.windll.kernel32.SetProcessAffinityMask
    SetProcessAffinityMask.argtypes = [ctypes.wintypes.HANDLE, DWORD_PTR]
    SetProcessAffinityMask.restype  = bool

    GetCurrentProcess = ctypes.windll.kernel32.GetCurrentProcess
    GetCurrentProcess.argtypes = []
    GetCurrentProcess.restype  = ctypes.wintypes.HANDLE

    def set_cpu_affinity(affinity_list):
        """Set CPU affinity to CPUs listed (numbered 0...n-1)"""
        mask = 0
        for num in affinity_list:
            mask |= 2**num

        # Pseudohandle, doesn't need to be closed
        handle = GetCurrentProcess()
        ok = SetProcessAffinityMask(handle, mask)
        if not ok:
            raise RuntimeError("SetProcessAffinityMask failed")
else:
    try:
        import resource

        # POSIX
        if sys.platform == 'darwin':
            def get_maxrss():
                # OSX getrusage returns maxrss in bytes
                # https://developer.apple.com/library/mac/documentation/Darwin/Reference/ManPages/man2/getrusage.2.html
                return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        else:
            def get_maxrss():
                # Linux, *BSD return maxrss in kilobytes
                return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
    except ImportError:
        pass

    def set_cpu_affinity(affinity_list):
        """Set CPU affinity to CPUs listed (numbered 0...n-1)"""
        if hasattr(os, 'sched_setaffinity'):
            os.sched_setaffinity(0, affinity_list)
        else:
            import psutil
            p = psutil.Process()
            if hasattr(p, 'cpu_affinity'):
                p.cpu_affinity(affinity_list)


def recvall(sock, size):
    """
    Receive data of given size from a socket connection
    """
    data = b""
    while len(data) < size:
        s = sock.recv(size - len(data))
        data += s
        if not s:
            raise RuntimeError("did not receive data from socket "
                               "(size {}, got only {!r})".format(size, data))
    return data


def _get_attr(source, name, ignore_case=False):
    if ignore_case:
        attrs = [getattr(source, key) for key in dir(source)
                 if key.lower() == name.lower()]

        if len(attrs) > 1:
            raise ValueError(
                "{0} contains multiple {1} functions.".format(
                    source.__name__, name))
        elif len(attrs) == 1:
            return attrs[0]
        else:
            return None
    else:
        return getattr(source, name, None)


def _get_all_attrs(sources, name, ignore_case=False):
    for source in sources:
        val = _get_attr(source, name, ignore_case=ignore_case)
        if val is not None:
            yield val


def _get_first_attr(sources, name, default, ignore_case=False):
    for val in _get_all_attrs(sources, name, ignore_case=ignore_case):
        return val
    return default


def get_setup_cache_key(func):
    if func is None:
        return None

    module = inspect.getmodule(func)
    mname = ".".join(module.__name__.split('.', 1)[1:])
    if not mname:
        mname = inspect.getsourcefile(func)

    return '{0}:{1}'.format(mname, inspect.getsourcelines(func)[1])


def get_source_code(items):
    """
    Extract source code of given items, and concatenate and dedent it.
    """
    sources = []
    prev_class_name = None

    for func in items:

        # custom source
        if hasattr(func, 'pretty_source'):
            src = textwrap.dedent(func.pretty_source).lstrip()
        # original source
        else:
            try:
                lines, _ = inspect.getsourcelines(func)
            except TypeError:
                continue

            if not lines:
                continue

            src = "\n".join(line.rstrip() for line in lines)
            src = textwrap.dedent(src)

        class_name = None
        if inspect.ismethod(func):
            # Add class name
            if hasattr(func, 'im_class'):
                class_name = func.im_class.__name__
            elif hasattr(func, '__qualname__'):
                names = func.__qualname__.split('.')
                if len(names) > 1:
                    class_name = names[-2]

        if class_name and prev_class_name != class_name:
            src = "class {0}:\n    {1}".format(
                class_name, src.replace("\n", "\n    "))
        elif class_name:
            src = "    {1}".format(
                class_name, src.replace("\n", "\n    "))

        sources.append(src)
        prev_class_name = class_name

    return "\n\n".join(sources).rstrip()


def _get_sourceline_info(obj, basedir):
    try:
        fn = inspect.getsourcefile(obj)
        fn = os.path.relpath(fn, basedir)
        _, lineno = inspect.getsourcelines(obj)
        return " in {!s}:{!s}".format(fn, lineno)
    except Exception:
        return ""


def check_num_args(root, benchmark_name, func, min_num_args, max_num_args=None):
    if max_num_args is None:
        max_num_args = min_num_args

    try:
        if sys.version_info[0] >= 3:
            info = inspect.getfullargspec(func)
        else:
            info = inspect.getargspec(func)
    except Exception as exc:
        print("{!s}: failed to check ({!r}{!s}): {!s}".format(
            benchmark_name, func, _get_sourceline_info(func, root), exc))
        return True

    max_args = len(info.args)

    if inspect.ismethod(func):
        max_args -= 1

    if info.defaults is not None:
        min_args = max_args - len(info.defaults)
    else:
        min_args = max_args

    if info.varargs is not None:
        max_args = float('inf')

    ok = (min_args <= max_num_args) and (min_num_args <= max_args)
    if not ok:
        if min_args == max_args:
            args_str = min_args
        else:
            args_str = "{}-{}".format(min_args, max_args)
        if min_num_args == max_num_args:
            num_args_str = min_num_args
        else:
            num_args_str = "{}-{}".format(min_num_args, max_num_args)
        print("{!s}: wrong number of arguments (for {!r}{!s}): expected {}, has {}".format(
            benchmark_name, func, _get_sourceline_info(func, root),
            num_args_str, args_str))

    return ok


def _repr_no_address(obj):
    result = repr(obj)
    address_regex = re.compile(r'^(<.*) at (0x[\da-fA-F]*)(>)$')
    match = address_regex.match(result)
    if match:
        suspected_address = match.group(2)
        # Double check this is the actual address
        default_result = object.__repr__(obj)
        match2 = address_regex.match(default_result)
        if match2:
            known_address = match2.group(2)
            if known_address == suspected_address:
                result = match.group(1) + match.group(3)

    return result


class Benchmark(object):
    """
    Represents a single benchmark.
    """
    # The regex of the name of function or method to be considered as
    # this type of benchmark.  The default in the base class, will
    # match nothing.
    name_regex = re.compile('^$')

    def __init__(self, name, func, attr_sources):
        self.name = name
        self.func = func
        self.pretty_name = getattr(func, "pretty_name", None)
        self._attr_sources = attr_sources
        self._setups = list(_get_all_attrs(attr_sources, 'setup', True))[::-1]
        self._teardowns = list(_get_all_attrs(attr_sources, 'teardown', True))
        self._setup_cache = _get_first_attr(attr_sources, 'setup_cache', None)
        self.setup_cache_key = get_setup_cache_key(self._setup_cache)
        self.setup_cache_timeout = _get_first_attr([self._setup_cache], "timeout", None)
        self.timeout = _get_first_attr(attr_sources, "timeout", 60.0)
        self.code = get_source_code([self.func] + self._setups + [self._setup_cache])
        if sys.version_info[0] >= 3:
            code_text = self.code.encode('utf-8')
        else:
            code_text = self.code
        code_hash = sha256(code_text).hexdigest()
        self.version = str(_get_first_attr(attr_sources, "version", code_hash))
        self.type = "base"
        self.unit = "unit"

        self._redo_setup_next = False

        self._params = _get_first_attr(attr_sources, "params", [])
        self.param_names = _get_first_attr(attr_sources, "param_names", [])
        self._current_params = ()

        # Enforce params format
        try:
            self.param_names = [str(x) for x in list(self.param_names)]
        except ValueError:
            raise ValueError("%s.param_names is not a list of strings" % (name,))

        try:
            self._params = list(self._params)
        except ValueError:
            raise ValueError("%s.params is not a list" % (name,))

        if self._params and not isinstance(self._params[0], (tuple, list)):
            # Accept a single list for one parameter only
            self._params = [self._params]
        else:
            self._params = [[item for item in entry] for entry in self._params]

        if len(self.param_names) != len(self._params):
            self.param_names = self.param_names[:len(self._params)]
            self.param_names += ['param%d' % (k+1,) for k in range(len(self.param_names),
                                                                   len(self._params))]

        # Exported parameter representations
        self.params = [[_repr_no_address(item) for item in entry] for entry in self._params]
        for i, param in enumerate(self.params):
            if len(param) != len(set(param)):
                counter = Counter(param)
                dupe_dict = {name: 0 for name, count in counter.items() if count > 1}
                for j in range(len(param)):
                    name = param[j]
                    if name in dupe_dict:
                        param[j] = name + ' ({})'.format(dupe_dict[name])
                        dupe_dict[name] += 1
                self.params[i] = param

    def set_param_idx(self, param_idx):
        try:
            self._current_params, = itertools.islice(
                itertools.product(*self._params),
                param_idx, param_idx + 1)
        except ValueError:
            raise ValueError(
                "Invalid benchmark parameter permutation index: %r" % (param_idx,))

    def insert_param(self, param):
        """
        Insert a parameter at the front of the parameter list.
        """
        self._current_params = tuple([param] + list(self._current_params))

    def __repr__(self):
        return '<{0} {1}>'.format(self.__class__.__name__, self.name)

    def check(self, root):
        # Check call syntax (number of arguments only...)
        ok = True

        if self._params:
            self.set_param_idx(0)

        min_num_args = len(self._current_params)
        max_num_args = min_num_args

        if self.setup_cache_key is not None:
            ok = ok and check_num_args(root, self.name + ": setup_cache",
                                       self._setup_cache, 0)
            max_num_args += 1

        for setup in self._setups:
            ok = ok and check_num_args(root, self.name + ": setup",
                                       setup, min_num_args, max_num_args)

        ok = ok and check_num_args(root, self.name + ": call",
                                   self.func, min_num_args, max_num_args)

        for teardown in self._teardowns:
            ok = ok and check_num_args(root, self.name + ": teardown",
                                       teardown, min_num_args, max_num_args)

        return ok

    def do_setup(self):
        try:
            for setup in self._setups:
                setup(*self._current_params)
        except NotImplementedError as e:
            # allow skipping test
            print("asv: skipped: {!r} ".format(e))
            return True
        return False

    def redo_setup(self):
        if not self._redo_setup_next:
            self._redo_setup_next = True
            return
        self.do_teardown()
        self.do_setup()

    def do_teardown(self):
        for teardown in self._teardowns:
            teardown(*self._current_params)

    def do_setup_cache(self):
        if self._setup_cache is not None:
            return self._setup_cache()

    def do_run(self):
        return self.run(*self._current_params)

    def do_profile(self, filename=None):
        def method_caller():
            run(*params)

        if profile is None:
            raise RuntimeError("cProfile could not be imported")

        if filename is not None:
            if hasattr(method_caller, 'func_code'):
                code = method_caller.func_code
            else:
                code = method_caller.__code__

            self.redo_setup()

            profile.runctx(
                code, {'run': self.func, 'params': self._current_params},
                {}, filename)


class TimeBenchmark(Benchmark):
    """
    Represents a single benchmark for timing.
    """
    name_regex = re.compile(
        '^(Time[A-Z_].+)|(time_.+)$')

    def __init__(self, name, func, attr_sources):
        Benchmark.__init__(self, name, func, attr_sources)
        self.type = "time"
        self.unit = "seconds"
        self._attr_sources = attr_sources
        old = int(_get_first_attr(self._attr_sources, 'processes', 2))  # backward compat.
        self.rounds = int(_get_first_attr(self._attr_sources, 'rounds', old))
        self._load_vars()

    def _load_vars(self):
        self.repeat = _get_first_attr(self._attr_sources, 'repeat', 0)
        self.min_run_count = _get_first_attr(self._attr_sources, 'min_run_count', 2)
        self.number = int(_get_first_attr(self._attr_sources, 'number', 0))
        self.sample_time = _get_first_attr(self._attr_sources, 'sample_time', 0.01)
        self.warmup_time = _get_first_attr(self._attr_sources, 'warmup_time', -1)
        self.timer = _get_first_attr(self._attr_sources, 'timer', wall_timer)

    def do_setup(self):
        result = Benchmark.do_setup(self)
        # For parameterized tests, setup() is allowed to change these
        self._load_vars()
        return result

    def _get_timer(self, *param):
        if param:
            func = lambda: self.func(*param)
        else:
            func = self.func

        timer = timeit.Timer(
            stmt=func,
            setup=self.redo_setup,
            timer=self.timer)

        return timer

    def run(self, *param):
        warmup_time = self.warmup_time
        if warmup_time < 0:
            if '__pypy__' in sys.modules:
                warmup_time = 1.0
            else:
                # Transient effects exist also on CPython, e.g. from
                # OS scheduling
                warmup_time = 0.1

        timer = self._get_timer(*param)

        try:
            min_repeat, max_repeat, max_time = self.repeat
        except (ValueError, TypeError):
            if self.repeat == 0:
                min_repeat = 1
                max_repeat = 10
                max_time = 20.0
                if self.rounds > 1:
                    max_repeat //= 2
                    max_time /= 2.0
            else:
                min_repeat = self.repeat
                max_repeat = self.repeat
                max_time = self.timeout

        min_repeat = int(min_repeat)
        max_repeat = int(max_repeat)
        max_time = float(max_time)

        samples, number = self.benchmark_timing(timer, min_repeat, max_repeat,
                                                max_time=max_time,
                                                warmup_time=warmup_time,
                                                number=self.number,
                                                min_run_count=self.min_run_count)

        samples = [s/number for s in samples]
        return {'samples': samples, 'number': number}

    def benchmark_timing(self, timer, min_repeat, max_repeat, max_time, warmup_time,
                         number, min_run_count):

        sample_time = self.sample_time
        start_time = wall_timer()
        run_count = 0

        samples = []

        def too_slow(num_samples):
            # stop taking samples if limits exceeded
            if run_count < min_run_count:
                return False
            if num_samples < min_repeat:
                return False
            return wall_timer() > start_time + warmup_time + max_time

        if number == 0:
            # Select number & warmup.
            #
            # This needs to be done at the same time, because the
            # benchmark timings at the beginning can be larger, and
            # lead to too small number being selected.
            number = 1
            while True:
                self._redo_setup_next = False
                start = wall_timer()
                timing = timer.timeit(number)
                wall_time = wall_timer() - start
                actual_timing = max(wall_time, timing)
                run_count += number

                if actual_timing >= sample_time:
                    if wall_timer() > start_time + warmup_time:
                        break
                else:
                    try:
                        p = min(10.0, max(1.1, sample_time/actual_timing))
                    except ZeroDivisionError:
                        p = 10.0
                    number = max(number + 1, int(p * number))

            if too_slow(1):
                return [timing], number
        elif warmup_time > 0:
            # Warmup
            while True:
                self._redo_setup_next = False
                timing = timer.timeit(number)
                run_count += number
                if wall_timer() >= start_time + warmup_time:
                    break

            if too_slow(1):
                return [timing], number

        # Collect samples
        while len(samples) < max_repeat:
            timing = timer.timeit(number)
            run_count += number
            samples.append(timing)

            if too_slow(len(samples)):
                break

        return samples, number


class _SeparateProcessTimer(object):
    subprocess_tmpl = textwrap.dedent('''
        from __future__ import print_function
        from timeit import timeit, default_timer as timer
        print(repr(timeit(stmt="""{stmt}""", setup="""{setup}""", number={number}, timer=timer)))
    ''').strip()

    def __init__(self, func):
        self.func = func

    def timeit(self, number):
        stmt = self.func()
        if isinstance(stmt, tuple):
            stmt, setup = stmt
        else:
            setup = ""
        stmt = textwrap.dedent(stmt)
        setup = textwrap.dedent(setup)
        stmt = stmt.replace(r'"""', r'\"\"\"')
        setup = setup.replace(r'"""', r'\"\"\"')

        code = self.subprocess_tmpl.format(stmt=stmt, setup=setup, number=number)

        res = subprocess.check_output([sys.executable, "-c", code])
        return float(res.strip())


class TimerawBenchmark(TimeBenchmark):
    """
    Represents a benchmark for tracking timing benchmarks run once in
    a separate process.
    """
    name_regex = re.compile(
        '^(Timeraw[A-Z_].+)|(timeraw_.+)$')

    def _load_vars(self):
        TimeBenchmark._load_vars(self)
        self.number = int(_get_first_attr(self._attr_sources, 'number', 1))
        del self.timer

    def _get_timer(self, *param):
        if param:
            func = lambda: self.func(*param)
        else:
            func = self.func
        return _SeparateProcessTimer(func)

    def do_profile(self, filename=None):
        raise ValueError("Raw timing benchmarks cannot be profiled")


class MemBenchmark(Benchmark):
    """
    Represents a single benchmark for tracking the memory consumption
    of an object.
    """
    name_regex = re.compile(
        '^(Mem[A-Z_].+)|(mem_.+)$')

    def __init__(self, name, func, attr_sources):
        Benchmark.__init__(self, name, func, attr_sources)
        self.type = "memory"
        self.unit = "bytes"

    def run(self, *param):
        # We can't import asizeof directly, because we haven't loaded
        # the asv package in the benchmarking process.
        path = os.path.join(
            os.path.dirname(__file__), 'extern', 'asizeof.py')
        if sys.version_info[0] >= 3:
            asizeof = importlib.machinery.SourceFileLoader('asizeof', path).load_module()
        else:
            asizeof = imp.load_source('asizeof', path)

        obj = self.func(*param)

        sizeof2 = asizeof.asizeof([obj, obj])
        sizeofcopy = asizeof.asizeof([obj, copy.copy(obj)])

        return sizeofcopy - sizeof2


class PeakMemBenchmark(Benchmark):
    """
    Represents a single benchmark for tracking the peak memory consumption
    of the whole program.
    """
    name_regex = re.compile(
        '^(PeakMem[A-Z_].+)|(peakmem_.+)$')

    def __init__(self, name, func, attr_sources):
        Benchmark.__init__(self, name, func, attr_sources)
        self.type = "peakmemory"
        self.unit = "bytes"

    def run(self, *param):
        self.func(*param)
        return get_maxrss()


class TrackBenchmark(Benchmark):
    """
    Represents a single benchmark for tracking an arbitrary value.
    """
    name_regex = re.compile(
        '^(Track[A-Z_].+)|(track_.+)$')

    def __init__(self, name, func, attr_sources):
        Benchmark.__init__(self, name, func, attr_sources)
        self.type = _get_first_attr(attr_sources, "type", "track")
        self.unit = _get_first_attr(attr_sources, "unit", "unit")

    def run(self, *param):
        return self.func(*param)


# TODO: Support the creation of custom benchmark types


benchmark_types = [
    TimerawBenchmark, TimeBenchmark, MemBenchmark, PeakMemBenchmark, TrackBenchmark
]


class SpecificImporter(object):
    """
    Module importer that only allows loading a given module from the
    given path.

    Using this enables importing the asv benchmark suite without
    adding its parent directory to sys.path. The parent directory can
    in principle contain anything, including some version of the
    project module (common situation if asv.conf.json is on project
    repository top level).
    """

    def __init__(self, name, root):
        self._name = name
        self._root = root

    if sys.version_info[0] >= 3:
        def find_spec(self, fullname, path, target):
            if fullname == self._name:
                if path is not None:
                    raise ValueError()
                finder = importlib.machinery.PathFinder()
                return finder.find_spec(fullname, [self._root], target)
            return None
    else:
        def find_module(self, fullname, path=None):
            if fullname == self._name:
                return self
            return None

        def load_module(self, fullname):
            file, pathname, desc = imp.find_module(fullname, [self._root])
            return imp.load_module(fullname, file, pathname, desc)


def update_sys_path(root):
    sys.meta_path.insert(0, SpecificImporter(os.path.basename(root),
                                             os.path.dirname(root)))


def _get_benchmark(attr_name, module, klass, func):
    try:
        name = func.benchmark_name
    except AttributeError:
        name = None
        search = attr_name
    else:
        search = name.split('.')[-1]

    for cls in benchmark_types:
        if cls.name_regex.match(search):
            break
    else:
        return
    # relative to benchmark_dir
    mname_parts = module.__name__.split('.', 1)[1:]
    if klass is None:
        if name is None:
            name = ".".join(mname_parts + [func.__name__])
        sources = [func, module]
    else:
        instance = klass()
        func = getattr(instance, attr_name)
        if name is None:
            name = ".".join(mname_parts + [klass.__name__, attr_name])
        sources = [func, instance, module]
    return cls(name, func, sources)


def disc_modules(module_name, ignore_import_errors=False):
    """
    Recursively import a module and all sub-modules in the package

    Yields
    ------
    module
        Imported module in the package tree

    """
    if not ignore_import_errors:
        module = import_module(module_name)
    else:
        try:
            module = import_module(module_name)
        except BaseException:
            traceback.print_exc()
            return

    # Exclude sourceless .pyc/.pyo left around (py3 __pycache__
    # behaves sensibly, so workaround only for py2)
    if sys.version_info[0] == 2 and not inspect.getsourcefile(module):
        return

    yield module

    if getattr(module, '__path__', None):
        for _, name, _ in pkgutil.iter_modules(module.__path__, module_name + '.'):
            for item in disc_modules(name, ignore_import_errors=ignore_import_errors):
                yield item


def disc_benchmarks(root, ignore_import_errors=False):
    """
    Discover all benchmarks in a given directory tree, yielding Benchmark
    objects

    For each class definition, looks for any methods with a
    special name.

    For each free function, yields all functions with a special
    name.
    """

    root_name = os.path.basename(root)

    for module in disc_modules(root_name, ignore_import_errors=ignore_import_errors):
        for attr_name, module_attr in (
            (k, v) for k, v in module.__dict__.items()
            if not k.startswith('_')
        ):
            if (inspect.isclass(module_attr) and
                    not inspect.isabstract(module_attr)):
                for name, class_attr in inspect.getmembers(module_attr):
                    if (inspect.isfunction(class_attr) or
                            inspect.ismethod(class_attr)):
                        benchmark = _get_benchmark(name, module, module_attr,
                                                   class_attr)
                        if benchmark is not None:
                            yield benchmark
            elif inspect.isfunction(module_attr):
                benchmark = _get_benchmark(attr_name, module, None, module_attr)
                if benchmark is not None:
                    yield benchmark


def get_benchmark_from_name(root, name, extra_params=None):
    """
    Create a benchmark from a fully-qualified benchmark name.

    Parameters
    ----------
    root : str
        Path to the root of a benchmark suite.

    name : str
        Fully-qualified name to a specific benchmark.
    """

    if '-' in name:
        try:
            name, param_idx = name.split('-', 1)
            param_idx = int(param_idx)
        except ValueError:
            raise ValueError("Benchmark id %r is invalid" % (name,))
    else:
        param_idx = None

    update_sys_path(root)
    benchmark = None

    # try to directly import benchmark function by guessing its import module
    # name
    parts = name.split('.')
    for i in [1, 2]:
        path = os.path.join(root, *parts[:-i]) + '.py'
        if not os.path.isfile(path):
            continue
        modname = '.'.join([os.path.basename(root)] + parts[:-i])
        module = import_module(modname)
        try:
            module_attr = getattr(module, parts[-i])
        except AttributeError:
            break
        if i == 1 and inspect.isfunction(module_attr):
            benchmark = _get_benchmark(parts[-i], module, None, module_attr)
            break
        elif i == 2 and inspect.isclass(module_attr):
            try:
                class_attr = getattr(module_attr, parts[-1])
            except AttributeError:
                break
            if (inspect.isfunction(class_attr) or
                    inspect.ismethod(class_attr)):
                benchmark = _get_benchmark(parts[-1], module, module_attr,
                                           class_attr)
                break

    if benchmark is None:
        for benchmark in disc_benchmarks(root):
            if benchmark.name == name:
                break
        else:
            raise ValueError(
                "Could not find benchmark '{0}'".format(name))

    if param_idx is not None:
        benchmark.set_param_idx(param_idx)

    if extra_params:
        class ExtraBenchmarkAttrs:
            pass
        for key, value in extra_params.items():
            setattr(ExtraBenchmarkAttrs, key, value)
        benchmark._attr_sources.insert(0, ExtraBenchmarkAttrs)

    return benchmark


def list_benchmarks(root, fp):
    """
    List all of the discovered benchmarks to fp as JSON.
    """
    update_sys_path(root)

    # Streaming of JSON back out to the master process

    fp.write('[')
    first = True
    for benchmark in disc_benchmarks(root):
        if not first:
            fp.write(', ')
        clean = dict(
            (k, v) for (k, v) in benchmark.__dict__.items()
            if isinstance(v, (str, int, float, list, dict, bool)) and not
               k.startswith('_'))
        json.dump(clean, fp, skipkeys=True)
        first = False
    fp.write(']')


def main_discover(args):
    benchmark_dir, result_file = args
    with open(result_file, 'w') as fp:
        list_benchmarks(benchmark_dir, fp)


def main_check(args):
    benchmark_dir, = args

    update_sys_path(benchmark_dir)

    ok = True
    for benchmark in disc_benchmarks(benchmark_dir):
        ok = ok and benchmark.check(benchmark_dir)

    sys.exit(0 if ok else 1)


def set_cpu_affinity_from_params(extra_params):
    affinity_list = extra_params.get('cpu_affinity', None)
    if affinity_list is not None:
        try:
            set_cpu_affinity(affinity_list)
        except BaseException as exc:
            print("asv: setting cpu affinity {!r} failed: {!r}".format(
                affinity_list, exc))


def main_setup_cache(args):
    (benchmark_dir, benchmark_id, params_str) = args

    extra_params = json.loads(params_str)

    set_cpu_affinity_from_params(extra_params)

    benchmark = get_benchmark_from_name(benchmark_dir, benchmark_id)
    cache = benchmark.do_setup_cache()
    with open("cache.pickle", "wb") as fd:
        pickle.dump(cache, fd)


def main_run(args):
    (benchmark_dir, benchmark_id, params_str, profile_path, result_file) = args

    extra_params = json.loads(params_str)

    set_cpu_affinity_from_params(extra_params)
    extra_params.pop('cpu_affinity', None)

    if profile_path == 'None':
        profile_path = None

    benchmark = get_benchmark_from_name(
        benchmark_dir, benchmark_id, extra_params=extra_params)

    if benchmark.setup_cache_key is not None:
        with open("cache.pickle", "rb") as fd:
            cache = pickle.load(fd)
        if cache is not None:
            benchmark.insert_param(cache)

    skip = benchmark.do_setup()

    try:
        if skip:
            result = float('nan')
        else:
            result = benchmark.do_run()
            if profile_path is not None:
                benchmark.do_profile(profile_path)
    finally:
        benchmark.do_teardown()

    # Write the output value
    with open(result_file, 'w') as fp:
        json.dump(result, fp)


@contextlib.contextmanager
def posix_redirect_output(filename=None, permanent=True):
    """
    Redirect stdout/stderr to a file, using posix dup2.
    """
    sys.stdout.flush()
    sys.stderr.flush()

    stdout_fd = sys.stdout.fileno()
    stderr_fd = sys.stderr.fileno()

    if not permanent:
        stdout_fd_copy = os.dup(stdout_fd)
        stderr_fd_copy = os.dup(stderr_fd)

    if filename is None:
        out_fd, filename = tempfile.mkstemp()
    else:
        out_fd = os.open(filename, os.O_WRONLY | os.O_CREAT | os.O_TRUNC)

    try:
        # Redirect stdout and stderr to file
        os.dup2(out_fd, stdout_fd)
        os.dup2(out_fd, stderr_fd)

        yield filename
    finally:
        sys.stdout.flush()
        sys.stderr.flush()
        os.close(out_fd)

        if not permanent:
            os.dup2(stdout_fd_copy, stdout_fd)
            os.dup2(stderr_fd_copy, stderr_fd)
            os.close(stdout_fd_copy)
            os.close(stderr_fd_copy)


def main_run_server(args):
    import io
    import signal
    import socket

    benchmark_dir, socket_name, = args

    update_sys_path(benchmark_dir)

    # Socket I/O
    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    s.bind(socket_name)
    s.listen(1)

    # Read and act on commands from socket
    while True:
        stdout_file = None

        try:
            conn, addr = s.accept()
        except KeyboardInterrupt:
            break

        try:
            fd, stdout_file = tempfile.mkstemp()
            os.close(fd)

            # Read command
            read_size, = struct.unpack('<Q', recvall(conn, 8))
            command_text = recvall(conn, read_size)
            if sys.version_info[0] >= 3:
                command_text = command_text.decode('utf-8')

            # Parse command
            command = json.loads(command_text)
            action = command.pop('action')

            if action == 'quit':
                break
            elif action == 'preimport':
                # Import benchmark suite before forking.
                # Capture I/O to a file during import.
                with posix_redirect_output(stdout_file, permanent=False):
                    for benchmark in disc_benchmarks(benchmark_dir, ignore_import_errors=True):
                        pass

                # Report result
                with io.open(stdout_file, 'r', errors='replace') as f:
                    out = f.read()
                out = json.dumps(out)
                if sys.version_info[0] >= 3:
                    out = out.encode('utf-8')
                conn.sendall(struct.pack('<Q', len(out)))
                conn.sendall(out)
                continue

            benchmark_id = command.pop('benchmark_id')
            params_str = command.pop('params_str')
            profile_path = command.pop('profile_path')
            result_file = command.pop('result_file')
            timeout = command.pop('timeout')
            cwd = command.pop('cwd')

            if command:
                raise RuntimeError('Command contained unknown data: {!r}'.format(command_text))

            # Spawn benchmark
            run_args = (benchmark_dir, benchmark_id, params_str, profile_path, result_file)
            pid = os.fork()
            if pid == 0:
                conn.close()
                sys.stdin.close()
                exitcode = 1
                try:
                    with posix_redirect_output(stdout_file, permanent=True):
                        try:
                            os.chdir(cwd)
                            main_run(run_args)
                            exitcode = 0
                        except BaseException as ec:
                            import traceback
                            traceback.print_exc()
                finally:
                    os._exit(exitcode)

            # Wait for results
            # (Poll in a loop is simplest --- also used by subprocess.py)
            start_time = wall_timer()
            is_timeout = False
            while True:
                res, status = os.waitpid(pid, os.WNOHANG)
                if res != 0:
                    break

                if timeout is not None and wall_timer() > start_time + timeout:
                    # Timeout
                    if is_timeout:
                        os.kill(pid, signal.SIGKILL)
                    else:
                        os.kill(pid, signal.SIGTERM)
                    is_timeout = True

                time.sleep(0.05)

            # Report result
            with io.open(stdout_file, 'r', errors='replace') as f:
                out = f.read()

            # Emulate subprocess
            if os.WIFSIGNALED(status):
                retcode = -os.WTERMSIG(status)
            elif os.WIFEXITED(status):
                retcode = os.WEXITSTATUS(status)
            elif os.WIFSTOPPED(status):
                retcode = -os.WSTOPSIG(status)
            else:
                # shouldn't happen, but fail silently
                retcode = -128

            info = {'out': out,
                    'errcode': -256 if is_timeout else retcode}

            result_text = json.dumps(info)
            if sys.version_info[0] >= 3:
                result_text = result_text.encode('utf-8')

            conn.sendall(struct.pack('<Q', len(result_text)))
            conn.sendall(result_text)
        except KeyboardInterrupt:
            break
        finally:
            conn.close()
            if stdout_file is not None:
                os.unlink(stdout_file)


def main_timing(argv):
    import argparse
    import asv.statistics
    import asv.util
    import asv.console

    if (_old_sys_path_head is not None and
        os.path.abspath(_old_sys_path_head) != os.path.abspath(os.path.dirname(__file__))):
        sys.path.insert(0, _old_sys_path_head)

    parser = argparse.ArgumentParser(usage="python -masv.benchmark timing [options] STATEMENT")
    parser.add_argument("--setup", action="store", default=(lambda: None))
    parser.add_argument("--number", action="store", type=int, default=0)
    parser.add_argument("--repeat", action="store", type=int, default=0)
    parser.add_argument("--timer", action="store", choices=("process_time", "perf_counter"),
                        default="perf_counter")
    parser.add_argument("--json", action="store_true")
    parser.add_argument("statement")
    args = parser.parse_args(argv)

    timer_func = {
        "process_time": process_time,
        "perf_counter": timeit.default_timer,
    }[args.timer]

    class AttrSource:
        pass

    attrs = AttrSource()
    attrs.repeat = args.repeat
    attrs.number = args.number
    attrs.timer = timer_func

    bench = TimeBenchmark("tmp", args.statement, [attrs])
    bench.redo_setup = args.setup
    result = bench.run()

    value, stats = asv.statistics.compute_stats(result['samples'], result['number'])
    formatted = asv.util.human_time(value, asv.statistics.get_err(value, stats))

    if not args.json:
        asv.console.color_print(formatted, 'red')
        asv.console.color_print(u"", 'default')
        asv.console.color_print(u"\n".join(u"{}: {}".format(k, v) for k, v in sorted(stats.items())), 'default')
        asv.console.color_print(u"samples: {}".format(result['samples']), 'default')
    else:
        json.dump({'result': value,
                   'samples': result['samples'],
                   'stats': stats}, sys.stdout)


def main_help(args):
    print(__doc__)


commands = {
    'discover': main_discover,
    'setup_cache': main_setup_cache,
    'run': main_run,
    'run_server': main_run_server,
    'check': main_check,
    'timing': main_timing,
    '-h': main_help,
    '--help': main_help,
}


def main():
    if len(sys.argv) < 2:
        main_help([])
        sys.exit(1)

    mode = sys.argv[1]
    args = sys.argv[2:]

    if mode in commands:
        commands[mode](args)
        sys.exit(0)
    else:
        sys.stderr.write("Unknown mode {0}\n".format(mode))
        sys.exit(1)

if __name__ == '__main__':
    main()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, unicode_literals, print_function

import os
import re
import itertools
import datetime
import six

from six.moves.urllib.parse import urlencode

from ..results import iter_results
from ..console import log
from ..publishing import OutputPublisher
from ..step_detect import detect_regressions

from .. import util
from .. import feed


class Regressions(OutputPublisher):
    name = "regressions"
    button_label = "Show regressions"
    description = "Display information about recent regressions"
    order = 3

    @classmethod
    def publish(cls, conf, repo, benchmarks, graphs, revisions):
        # Analyze the data in the graphs --- it's been cleaned up and
        # it's easier to work with than the results directly

        regressions = []
        revision_to_hash = dict((r, h) for h, r in six.iteritems(revisions))

        data_filter = _GraphDataFilter(conf, repo, revisions)

        all_params = graphs.get_params()

        for j, (file_name, graph) in enumerate(graphs):
            if 'summary' in graph.params:
                continue

            benchmark_name = os.path.basename(file_name)
            benchmark = benchmarks.get(benchmark_name)
            if not benchmark:
                continue

            log.dot()

            for graph_data in data_filter.get_graph_data(graph, benchmark):
                cls._process_regression(regressions, revision_to_hash, repo, all_params,
                                        graph_data, graph)

        cls._save(conf, {'regressions': regressions})
        cls._save_feed(conf, benchmarks, regressions, revisions, revision_to_hash)

    @classmethod
    def _process_regression(cls, regressions, revision_to_hash, repo, all_params,
                           graph_data, graph):
        j, entry_name, steps, threshold = graph_data

        last_v, best_v, jumps = detect_regressions(steps, threshold)

        if last_v is None:
            return

        # Select unique graph params
        graph_params = {}
        for name, value in six.iteritems(graph.params):
            if len(all_params[name]) > 1:
                graph_params[name] = value

        graph_path = graph.path + '.json'

        # Check which ranges are a single commit
        for k, jump in enumerate(jumps):
            commit_a = revision_to_hash[jump[0]]
            commit_b = revision_to_hash[jump[1]]
            spec = repo.get_range_spec(commit_a, commit_b)
            commits = repo.get_hashes_from_range(spec)
            if len(commits) == 1:
                jumps[k] = (None, jump[1], jump[2], jump[3])

        # Produce output
        regression = [entry_name, graph_path, graph_params, j, last_v, best_v, jumps]
        regressions.append(regression)

    @classmethod
    def _save(cls, conf, data):
        fn = os.path.join(conf.html_dir, 'regressions.json')
        util.write_json(fn, data, compact=True)

    @classmethod
    def _save_feed(cls, conf, benchmarks, data, revisions, revision_to_hash):
        """
        Save the results as an Atom feed
        """

        filename = os.path.join(conf.html_dir, 'regressions.xml')

        # Determine publication date as the date when the benchmark
        # was run --- if it is missing, use the date of the commit
        run_timestamps = {}
        revision_timestamps = {}
        for results in iter_results(conf.results_dir):
            if results.commit_hash not in revisions:
                # revisions could be filtered when specifying a range
                # in 'asv publish'
                continue
            revision = revisions[results.commit_hash]
            revision_timestamps[revision] = results.date

            # Time when the benchmark was run
            for benchmark_name, timestamp in six.iteritems(results.started_at):
                if timestamp is None:
                    continue
                key = (benchmark_name, revision)
                run_timestamps[key] = timestamp

            # Fallback to commit date
            for benchmark_name in results.get_result_keys(benchmarks):
                key = (benchmark_name, revision)
                run_timestamps.setdefault(key, results.date)

        # Generate feed entries
        entries = []

        for name, graph_path, graph_params, idx, last_value, best_value, jumps in data:
            if '(' in name:
                benchmark_name = name[:name.index('(')]
            else:
                benchmark_name = name

            benchmark = benchmarks[benchmark_name]

            if idx is not None:
                graph_params = dict(graph_params)

                # Add URL parameters
                param_values, = itertools.islice(itertools.product(*benchmark['params']),
                                                 idx, idx + 1)
                for k, v in zip(benchmark['param_names'], param_values):
                    graph_params['p-' + k] = v

            for rev1, rev2, value1, value2 in jumps:
                timestamps = (run_timestamps[benchmark_name, t] for t in (rev1, rev2) if t is not None)
                last_timestamp = max(timestamps)

                updated = datetime.datetime.fromtimestamp(last_timestamp/1000)

                params = dict(graph_params)

                if rev1 is None:
                    params['commits'] = '{0}'.format(revision_to_hash[rev2])
                else:
                    params['commits'] = '{0}-{1}'.format(revision_to_hash[rev1],
                                                         revision_to_hash[rev2])

                link = 'index.html#{0}?{1}'.format(benchmark_name, urlencode(params))

                try:
                    best_percentage = "{0:.2f}%".format(100 * (last_value - best_value) / best_value)
                except ZeroDivisionError:
                    best_percentage = "{0:.2g} units".format(last_value - best_value)

                try:
                    percentage = "{0:.2f}%".format(100 * (value2 - value1) / value1)
                except ZeroDivisionError:
                    percentage = "{0:.2g} units".format(value2 - value1)

                jump_date = datetime.datetime.fromtimestamp(revision_timestamps[rev2]/1000)
                jump_date_str = jump_date.strftime('%Y-%m-%d %H:%M:%S')

                if rev1 is not None:
                    commit_a = revision_to_hash[rev1]
                    commit_b = revision_to_hash[rev2]
                    if 'github.com' in conf.show_commit_url:
                        commit_url = conf.show_commit_url + '../compare/' + commit_a + "..." + commit_b
                    else:
                        commit_url = conf.show_commit_url + commit_a
                    commit_ref = 'in commits <a href="{0}">{1}...{2}</a>'.format(commit_url,
                                                                                 commit_a[:8],
                                                                                 commit_b[:8])
                else:
                    commit_a = revision_to_hash[rev2]
                    commit_url = conf.show_commit_url + commit_a
                    commit_ref = 'in commit <a href="{0}">{1}</a>'.format(commit_url, commit_a[:8])

                unit = benchmark.get('unit', '')
                best_value_str = util.human_value(best_value, unit)
                last_value_str = util.human_value(last_value, unit)
                value1_str = util.human_value(value1, unit)
                value2_str = util.human_value(value2, unit)

                title = "{percentage} {name}".format(**locals())
                summary = """
                <a href="{link}">{percentage} regression</a> on {jump_date_str} {commit_ref}.<br>
                New value: {value2_str}, old value: {value1_str}.<br>
                Latest value: {last_value_str} ({best_percentage} worse than best value {best_value_str}).
                """.format(**locals()).strip()

                # Information that uniquely identifies a regression
                # --- if the values and the texts change on later
                # runs, feed readers should is identify the regression
                # as the same one, as long as the benchmark name and
                # commits match.
                id_context = [name, revision_to_hash.get(rev1, ""), revision_to_hash.get(rev2, "")]
                id_date = util.js_timestamp_to_datetime(revision_timestamps[rev2])

                entries.append(feed.FeedEntry(title, updated, link, summary, id_context, id_date))

        entries.sort(key=lambda x: x.updated, reverse=True)

        feed.write_atom(filename, entries,
                        title='{0} performance regressions'.format(conf.project),
                        author='Airspeed Velocity',
                        address='{0}.asv'.format(conf.project))


class _GraphDataFilter(object):
    """
    Obtain data sets from graphs, following configuration settings.
    """

    def __init__(self, conf, repo, revisions):
        self.conf = conf
        self.repo = repo
        self.revisions = revisions
        self._start_revisions = {}

    def get_graph_data(self, graph, benchmark):
        """
        Iterator over graph data sets

        Yields
        ------
        param_idx
            Flat index to parameter permutations for parameterized benchmarks.
            None if benchmark is not parameterized.
        entry_name
            Name for the data set. If benchmark is non-parameterized, this is the
            benchmark name.
        steps
            Steps to consider in regression detection.
        threshold
            User-specified threshold for regression detection.

        """
        if benchmark.get('params'):
            param_iter = enumerate(zip(itertools.product(*benchmark['params']),
                                           graph.get_steps()))
        else:
            param_iter = [(None, (None, graph.get_steps()))]

        for j, (param, steps) in param_iter:
            if param is None:
                entry_name = benchmark['name']
            else:
                entry_name = benchmark['name'] + '({0})'.format(', '.join(param))

            start_revision = self._get_start_revision(graph, benchmark, entry_name)
            threshold = self._get_threshold(graph, benchmark, entry_name)

            if start_revision is None:
                # Skip detection
                continue

            steps = [step for step in steps if step[1] >= start_revision]

            yield j, entry_name, steps, threshold

    def _get_start_revision(self, graph, benchmark, entry_name):
        """
        Compute the first revision allowed by asv.conf.json.

        Revisions correspond to linearized commit history and the
        regression detection runs on this order --- the starting commit
        thus corresponds to a specific starting revision.
        """
        start_revision = min(six.itervalues(self.revisions))

        if graph.params.get('branch'):
            branch_suffix = '@' + graph.params.get('branch')
        else:
            branch_suffix = ''

        for regex, start_commit in six.iteritems(self.conf.regressions_first_commits):
            if re.match(regex, entry_name + branch_suffix):
                if start_commit is None:
                    # Disable regression detection completely
                    return None

                if self.conf.branches == [None]:
                    key = (start_commit, None)
                else:
                    key = (start_commit, graph.params.get('branch'))

                if key not in self._start_revisions:
                    spec = self.repo.get_new_range_spec(*key)
                    start_hash = self.repo.get_hash_from_name(start_commit)

                    for commit in [start_hash] + self.repo.get_hashes_from_range(spec):
                        rev = self.revisions.get(commit)
                        if rev is not None:
                            self._start_revisions[key] = rev
                            break
                    else:
                        # Commit not found in the branch --- warn and ignore.
                        log.warning(("Commit {0} specified in `regressions_first_commits` "
                                  "not found in branch").format(start_commit))
                        self._start_revisions[key] = -1

                start_revision = max(start_revision, self._start_revisions[key] + 1)

        return start_revision

    def _get_threshold(self, graph, benchmark, entry_name):
        """
        Compute the regression threshold in asv.conf.json.
        """
        if graph.params.get('branch'):
            branch_suffix = '@' + graph.params.get('branch')
        else:
            branch_suffix = ''

        max_threshold = None

        for regex, threshold in six.iteritems(self.conf.regressions_thresholds):
            if re.match(regex, entry_name + branch_suffix):
                try:
                    threshold = float(threshold)
                except ValueError:
                    raise util.UserError("Non-float threshold in asv.conf.json: {!r}".format(threshold))

                if max_threshold is None:
                    max_threshold = threshold
                else:
                    max_threshold = max(threshold, max_threshold)

        if max_threshold is None:
            max_threshold = 0.05

        return max_threshold



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import six

from fnmatch import fnmatchcase
import sys

from . import Command
from .. import console
from ..console import log
from ..results import iter_results
from . import util


class Rm(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "rm", help="Remove results from the database",
            description="""
            Removes entries from the results database.
            """)

        parser.add_argument(
            'patterns', nargs='+',
            help="""Pattern(s) to match, each of the form X=Y.  X may
            be one of "benchmark", "commit_hash", "python" or any of
            the machine or environment params.  Y is a case-sensitive
            glob pattern.""")
        parser.add_argument(
            "-y", action="store_true",
            help="""Don't prompt for confirmation.""")

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args):
        return cls.run(conf, args.patterns, args.y)

    @classmethod
    def run(cls, conf, patterns, y=True):
        global_patterns = {}
        single_benchmark = None
        files_to_remove = set()
        count = 0

        for pattern in patterns:
            parts = pattern.split('=', 1)
            if len(parts) != 2:
                raise util.UserError("Invalid pattern '{0}'".format(pattern))

            if parts[0] == 'benchmark':
                if single_benchmark is not None:
                    raise util.UserError("'benchmark' appears more than once")
                single_benchmark = parts[1]
            else:
                if parts[0] in global_patterns:
                    raise util.UserError(
                        "'{0}' appears more than once".format(parts[0]))
                global_patterns[parts[0]] = parts[1]

        for result in iter_results(conf.results_dir):
            found = True
            for key, val in six.iteritems(global_patterns):
                if key == 'commit_hash':
                    if not util.hash_equal(result.commit_hash, val):
                        found = False
                        break
                elif key == 'python':
                    if not fnmatchcase(result.env.python, val):
                        found = False
                        break
                else:
                    if not fnmatchcase(result.params.get(key), val):
                        found = False
                        break

            if not found:
                continue

            if single_benchmark is not None:
                found = False
                for benchmark in list(result.get_all_result_keys()):
                    if fnmatchcase(benchmark, single_benchmark):
                        count += 1
                        files_to_remove.add(result)
                        result.remove_result(benchmark)
            else:
                files_to_remove.add(result)

        if single_benchmark is not None:
            log.info("Removing {0} benchmarks in {1} files".format(
                count, len(files_to_remove)))
        else:
            log.info("Removing {0} files".format(len(files_to_remove)))

        if not y:
            do = console.get_answer_default("Perform operations", "n")
            if len(do) and do.lower()[0] != 'y':
                sys.exit(0)

        if single_benchmark is not None:
            for result in files_to_remove:
                result.save(conf.results_dir)
        else:
            for result in files_to_remove:
                result.rm(conf.results_dir)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from six.moves import SimpleHTTPServer, socketserver

import errno
import os
import random
import socket

from . import Command

from ..console import log
from .. import util


def random_ports(port, n):
    """Generate a list of n random ports near the given port.

    The first 5 ports will be sequential, and the remaining n-5 will be
    randomly selected in the range [port-2*n, port+2*n].
    """
    if port != 0:
        yield port
    else:
        port = 8080
        for i in range(min(5, n)):
            yield port + i
        for i in range(n-5):
            yield max(1, port + random.randint(-2*n, 2*n))


def create_httpd(handler_cls, port=0):
    # Create a server that allows address reuse
    class MyTCPServer(socketserver.ThreadingTCPServer):
        allow_reuse_address = True
        daemon_threads = True

    for port in random_ports(port, 5):
        try:
            httpd = MyTCPServer(("", port), handler_cls)
            base_url = "http://127.0.0.1:{0}/".format(port)
            break
        except socket.error as e:
            if e.errno == errno.EADDRINUSE:
                continue
            else:
                raise
    else:
        raise util.UserError("Failed to find an unused port for "
                             "serving web pages")

    return httpd, base_url


class Preview(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "preview",
            help="Preview the results using a local web server",
            description="Preview the results using a local web server")

        parser.add_argument("--port", "-p", type=int, default=0,
                            help="Port to run webserver on.  [8080]")
        parser.add_argument("--browser", "-b", action="store_true",
                            help="Open in webbrowser")
        parser.add_argument(
            '--html-dir', '-o', default=None, help=(
                "Optional output directory. Default is 'html_dir' "
                "from asv config"))

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args):
        if args.html_dir:
            conf.html_dir = args.html_dir
        return cls.run(conf=conf, port=args.port,
                       browser=args.browser)

    @classmethod
    def run(cls, conf, port=0, browser=False):
        os.chdir(conf.html_dir)

        class Handler(SimpleHTTPServer.SimpleHTTPRequestHandler):
            def translate_path(self, path):
                path = SimpleHTTPServer.SimpleHTTPRequestHandler.translate_path(
                    self, path)
                return util.long_path(path)

        httpd, base_url = create_httpd(Handler, port=port)

        log.info("Serving at {0}".format(base_url))

        if browser:
            import webbrowser
            webbrowser.open(base_url)

        log.info("Press ^C to abort\n")
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            return



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, unicode_literals, print_function

from distutils.version import LooseVersion
import sys
import re
import os

import six

from .. import environment
from ..console import log
from .. import util


WIN = (os.name == "nt")


class Virtualenv(environment.Environment):
    """
    Manage an environment using virtualenv.
    """
    tool_name = "virtualenv"

    def __init__(self, conf, python, requirements, tagged_env_vars):
        """
        Parameters
        ----------
        conf : Config instance

        python : str
            Version of Python.  Must be of the form "MAJOR.MINOR".

        executable : str
            Path to Python executable.

        requirements : dict
            Dictionary mapping a PyPI package name to a version
            identifier string.
        """
        executable = Virtualenv._find_python(python)
        if executable is None:
            raise environment.EnvironmentUnavailable(
                "No executable found for python {0}".format(python))

        self._executable = executable
        self._python = python
        self._requirements = requirements
        super(Virtualenv, self).__init__(conf,
                                         python,
                                         requirements,
                                         tagged_env_vars)

        try:
            import virtualenv
        except ImportError:
            raise environment.EnvironmentUnavailable(
                "virtualenv package not installed")

    @staticmethod
    def _find_python(python):
        """Find Python executable for the given Python version"""
        is_pypy = python.startswith("pypy")

        # Parse python specifier
        if is_pypy:
            executable = python
            if python == 'pypy':
                python_version = '2'
            else:
                python_version = python[4:]
        else:
            python_version = python
            executable = "python{0}".format(python_version)

        # Find Python executable on path
        try:
            return util.which(executable)
        except IOError:
            pass

        # Maybe the current one is correct?
        current_is_pypy = hasattr(sys, 'pypy_version_info')
        current_versions = ['{0[0]}'.format(sys.version_info),
                            '{0[0]}.{0[1]}'.format(sys.version_info)]

        if is_pypy == current_is_pypy and python_version in current_versions:
            return sys.executable

        return None

    @property
    def name(self):
        """
        Get a name to uniquely identify this environment.
        """
        python = self._python
        if self._python.startswith('pypy'):
            # get_env_name adds py-prefix
            python = python[2:]
        return environment.get_env_name(self.tool_name,
                                        python,
                                        self._requirements,
                                        self._tagged_env_vars)

    @classmethod
    def matches(self, python):
        if not (re.match(r'^[0-9].*$', python) or re.match(r'^pypy[0-9.]*$', python)):
            # The python name should be a version number, or pypy+number
            return False

        try:
            import virtualenv
        except ImportError:
            return False
        else:
            if LooseVersion(virtualenv.__version__) == LooseVersion('1.11.0'):
                log.warning(
                    "asv is not compatible with virtualenv 1.11 due to a bug in "
                    "setuptools.")
            if LooseVersion(virtualenv.__version__) < LooseVersion('1.10'):
                log.warning(
                    "If using virtualenv, it much be at least version 1.10")

        executable = Virtualenv._find_python(python)
        return executable is not None

    def _setup(self):
        """
        Setup the environment on disk using virtualenv.
        Then, all of the requirements are installed into
        it using `pip install`.
        """
        env = dict(os.environ)
        env.update(self.build_env_vars)

        log.info("Creating virtualenv for {0}".format(self.name))
        util.check_call([
            sys.executable,
            "-mvirtualenv",
            "-p",
            self._executable,
            self._path], env=env)

        log.info("Installing requirements for {0}".format(self.name))
        self._install_requirements()

    def _install_requirements(self):
        if sys.version_info[:2] == (3, 2):
            pip_args = ['install', '-v', 'wheel<0.29.0', 'pip<8']
        else:
            pip_args = ['install', '-v', 'wheel', 'pip>=8']

        env = dict(os.environ)
        env.update(self.build_env_vars)

        self._run_pip(pip_args, env=env)

        if self._requirements:
            args = ['install', '-v', '--upgrade']
            for key, val in six.iteritems(self._requirements):
                pkg = key
                if key.startswith('pip+'):
                    pkg = key[4:]

                if val:
                    args.append("{0}=={1}".format(pkg, val))
                else:
                    args.append(pkg)
            self._run_pip(args, timeout=self._install_timeout, env=env)

    def _run_pip(self, args, **kwargs):
        # Run pip via python -m pip, so that it works on Windows when
        # upgrading pip itself, and avoids shebang length limit on Linux
        return self.run_executable('python', ['-mpip'] + list(args), **kwargs)

    def run(self, args, **kwargs):
        log.debug("Running '{0}' in {1}".format(' '.join(args), self.name))
        return self.run_executable('python', args, **kwargs)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import shutil

from six.moves import input as raw_input

from . import Command
from ..console import log, color_print


class Quickstart(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "quickstart", help="Create a new benchmarking suite",
            description="Creates a new benchmarking suite")

        parser.add_argument(
            "--dest", "-d", default=".",
            help="The destination directory for the new benchmarking "
            "suite")

        grp = parser.add_mutually_exclusive_group()
        grp.add_argument(
            "--top-level", action="store_true", dest="top_level", default=None,
            help="Use layout suitable for putting the benchmark suite on "
            "the top level of the project's repository")
        grp.add_argument(
            "--no-top-level", action="store_false", dest="top_level", default=None,
            help="Use layout suitable for putting the benchmark suite in "
            "a separate repository")

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_args(cls, args):
        return cls.run(dest=args.dest, top_level=args.top_level)

    @classmethod
    def run(cls, dest=".", top_level=None):
        log.info("Setting up new Airspeed Velocity benchmark suite.")

        if top_level is None:
            log.flush()
            color_print("")
            color_print("Which of the following template layouts to use:")
            color_print("(1) benchmark suite at the top level of the project repository")
            color_print("(2) benchmark suite in a separate repository")
            color_print("")
            while True:
                answer = raw_input("Layout to use? [1/2] ")
                if answer.lower()[:1] == "1":
                    top_level = True
                    break
                elif answer.lower()[:1] == "2":
                    top_level = False
                    break
            color_print("")

        template_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), '..', 'template')
        for entry in os.listdir(template_path):
            path = os.path.join(template_path, entry)
            dest_path = os.path.join(dest, entry)
            if os.path.exists(dest_path):
                log.info("Template content already exists.")
                log.info("Edit asv.conf.json to continue.")
                return 1

        for entry in os.listdir(template_path):
            path = os.path.join(template_path, entry)
            dest_path = os.path.join(dest, entry)
            if os.path.isdir(path):
                shutil.copytree(path, os.path.join(dest, entry))
            elif os.path.isfile(path):
                shutil.copyfile(path, os.path.join(dest, entry))

        if top_level:
            conf_file = os.path.join(dest, 'asv.conf.json')

            with open(conf_file, 'r') as f:
                conf = f.read()

            reps = [('"repo": "",', '"repo": ".",'),
                    ('// "env_dir": "env",', '"env_dir": ".asv/env",'),
                    ('// "results_dir": "results",', '"results_dir": ".asv/results",'),
                    ('// "html_dir": "html",', '"html_dir": ".asv/html",')]
            for src, dst in reps:
                conf = conf.replace(src, dst)

            with open(conf_file, 'w') as f:
                f.write(conf)

        log.info("Edit asv.conf.json to get started.")



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Supports mercurial repositories for the benchmarked project.
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import sys
import re
try:
    import hglib
except ImportError as exc:
    hglib = None

from ..console import log
from ..repo import Repo, NoSuchNameError
from .. import util


class Hg(Repo):
    dvcs = "hg"
    _default_branch = "default"
    _encoding = "utf-8"

    def __init__(self, url, mirror_path):
        # TODO: shared repositories in Mercurial are only possible
        # through an extension, and it's not clear how to use those in
        # this context.  So here, we always make full clones for
        # each of the environments.

        self._path = os.path.abspath(mirror_path)
        self._pulled = False
        if hglib is None:
            raise ImportError("hglib")

        if self.is_local_repo(url):
            # Local repository, no need for mirror
            self._path = os.path.abspath(url)
            self._pulled = True
        elif not self.is_local_repo(self._path):
            if os.path.exists(self._path):
                self._raise_bad_mirror_error(self._path)

            # Clone is missing
            log.info("Cloning project")
            if url.startswith("hg+"):
                url = url[3:]

            # Mercurial branches are global, so there is no need for
            # an analog of git --mirror
            hglib.clone(self._encode_filename(url),
                        dest=self._encode_filename(self._path),
                        noupdate=True)

        self._repo = hglib.open(self._encode_filename(self._path))

    def __del__(self):
        if self._repo is not None:
            self._repo.close()
            self._repo = None

    def _decode(self, item):
        return item.decode(self._encoding)

    def _encode(self, item):
        return item.encode(self._encoding)

    def _encode_filename(self, filename):
        return filename.encode(sys.getfilesystemencoding())

    @classmethod
    def is_local_repo(cls, path):
        return (os.path.isdir(path) and
                os.path.isdir(os.path.join(path, '.hg')))

    @classmethod
    def url_match(cls, url):
        regexes = [
            r'^hg\+https?://.*$',
            r'^https?://.*?\.hg$',
            r'^ssh://hg@.*$']

        for regex in regexes:
            if re.match(regex, url):
                return True

        # Check for a local path
        if cls.is_local_repo(url):
            return True

        return False

    def get_range_spec(self, commit_a, commit_b):
        return '{0}::{1} and not {0}'.format(commit_a, commit_b, commit_a)

    def get_new_range_spec(self, latest_result, branch=None):
        return '{0}::{1}'.format(latest_result, self.get_branch_name(branch))

    def pull(self):
        # We assume the remote isn't updated during the run of asv
        # itself.
        if self._pulled:
            return

        log.info("Fetching recent changes")
        self._repo.pull()
        self._pulled = True

    def checkout(self, path, commit_hash):
        # Need to pull -- the copy is not updated automatically, since
        # the repository data is not shared

        def checkout_existing():
            with hglib.open(self._encode_filename(path)) as subrepo:
                subrepo.pull()
                subrepo.update(self._encode(commit_hash), clean=True)
                subrepo.rawcommand([b"--config",
                                    b"extensions.purge=",
                                    b"purge",
                                    b"--all"])

        if os.path.isdir(path):
            try:
                checkout_existing()
            except (hglib.error.CommandError, hglib.error.ServerError):
                # Remove and re-clone
                util.long_path_rmtree(path)

        if not os.path.isdir(path):
            hglib.clone(self._encode_filename(self._path),
                        dest=self._encode_filename(path))
            checkout_existing()

    def get_date(self, hash):
        # TODO: This works on Linux, but should be extended for other platforms
        rev = self._repo.log(self._encode(hash))[0]
        return int(rev.date.strftime("%s")) * 1000

    def get_hashes_from_range(self, range_spec, **kwargs):
        range_spec = self._encode("sort({0}, -rev)".format(range_spec))
        return [self._decode(rev.node) for rev in self._repo.log(range_spec, **kwargs)]

    def get_hash_from_name(self, name):
        if name is None:
            name = self.get_branch_name()
        try:
            return self._decode(self._repo.log(self._encode(name))[0].node)
        except hglib.error.CommandError as err:
            if b'unknown revision' in err.err:
                raise NoSuchNameError(name)
            raise

    def get_hash_from_parent(self, name):
        return self.get_hash_from_name('p1({0})'.format(name))

    def get_name_from_hash(self, commit):
        # XXX: implement
        return None

    def get_tags(self):
        tags = {}
        for item in self._repo.log(b"tag()"):
            tags[self._decode(item.tags)] = self._decode(item.node)
        return tags

    def get_date_from_name(self, name):
        return self.get_date(name)

    def get_branch_commits(self, branch):
        if self._repo.version >= (4, 5):
            query = "branch({0})"
        else:
            query = "ancestors({0})"
        return self.get_hashes_from_range(query.format(self.get_branch_name(branch)),
                                          followfirst=True)

    def get_revisions(self, commits):
        revisions = {}
        for i, item in enumerate(self._repo.log(b"all()")):
            node = self._decode(item.node)
            if node in commits:
                revisions[node] = i
        return revisions



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)


class ProfilerGui(object):
    """
    A base class to define a Profiler GUI that is available through
    the ``asv profile`` command.
    """
    name = None
    description = None

    @classmethod
    def is_available(cls):
        """
        Return `True` is the given GUI tool appears to be installed
        and available.
        """
        raise NotImplementedError()

    @classmethod
    def open_profiler_gui(cls, profiler_file):
        """
        Open the profiler GUI to display the results in the given
        profiler file.
        """
        raise NotImplementedError()
